/*
 * flash.c
 *
 * Copyright 2008,2009 Sony Corporation. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>
#include <linux/byteorder/swab.h>

#ifdef CONFIG_SONY_ICX

#define PHYS_FLASH_SIZE				0x01000000	/*  16 MB all		*/
#define PHYS_FLASH_BOOTSECT_SIZE	0x00008000	/*  32 KB sectors	*/
#define PHYS_FLASH_SECT_SIZE		0x00020000	/* 128 KB sectors	*/

#define	PHYS_FLASH_BOOTSECT_TOP		4
#define	PHYS_FLASH_BOOTSECT_BOTTOM	130

flash_info_t flash_info[CFG_MAX_FLASH_BANKS];	/* info for FLASH chips	   */

/* Board support for 1 or 2 flash devices */
#undef FLASH_PORT_WIDTH32
#define FLASH_PORT_WIDTH16

#ifdef FLASH_PORT_WIDTH16
#define FLASH_PORT_WIDTH		ushort
#define FLASH_PORT_WIDTHV		vu_short
#define SWAP(x)					__swab16(x)
#else
#define FLASH_PORT_WIDTH		ulong
#define FLASH_PORT_WIDTHV		vu_long
#define SWAP(x)					__swab32(x)
#endif

#define FPW		FLASH_PORT_WIDTH
#define FPWV	FLASH_PORT_WIDTHV

#define mb() __asm__ __volatile__ ("" : : : "memory")


/*-----------------------------------------------------------------------
 * Functions
 */
unsigned long	flash_init (void);
static ulong	flash_get_size (FPW * addr, flash_info_t * info);
static int		write_data (flash_info_t * info, ulong dest, FPW data);
static void		flash_get_offsets (ulong base, flash_info_t * info);
void inline		spin_wheel (ulong total, ulong cnt);
void			flash_print_info (flash_info_t * info);
void			flash_unprotect_sectors (FPWV * addr);
int				flash_erase (flash_info_t * info, int s_first, int s_last);
int				write_buff (flash_info_t * info, uchar * src, ulong addr, ulong cnt);
void			flash_unlock(flash_info_t * info);


/*-----------------------------------------------------------------------
 */
unsigned long flash_init (void)
{
	int i;
	ulong size = 0;
	ulong retnum = 0;

	for (i = 0; i < CFG_MAX_FLASH_BANKS; i++) {
		switch (i) {
		case 0:
			retnum = flash_get_size ((FPW *) PHYS_FLASH_1, &flash_info[i]);
			if (retnum == 0) {			/* PSRAM? */
				return 0;
			}
			flash_get_offsets (PHYS_FLASH_1, &flash_info[i]);
			/* to reset the lock bit */
			flash_unlock(&flash_info[i]);
			break;
		default:
			panic ("configured too many flash banks!\n");
			break;
		}
		size += flash_info[i].size;
	}

	return size;
}

/*-----------------------------------------------------------------------
 */
void flash_unlock(flash_info_t * info)
{
	return;
}

/*-----------------------------------------------------------------------
 */
static void flash_get_offsets (ulong base, flash_info_t * info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN) {
		return;
	}

	if ((info->flash_id & FLASH_VENDMASK) == FLASH_MAN_AMD) {
		if ((info->flash_id & FLASH_TYPEMASK) == FLASH_S29WS128P) {
			for (i = 0; i < info->sector_count; i++) {
				info->start[i] = base;
				info->protect[i] = 0;
				if ((i < PHYS_FLASH_BOOTSECT_TOP) || (i >= PHYS_FLASH_BOOTSECT_BOTTOM))
					base += PHYS_FLASH_BOOTSECT_SIZE;
				else
					base += PHYS_FLASH_SECT_SIZE;
			}
		}
	}
}

/*-----------------------------------------------------------------------
 */
void flash_print_info (flash_info_t * info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN) {
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id & FLASH_VENDMASK) {
	case FLASH_MAN_AMD:
		printf ("AMD ");
		break;
	default:
		printf ("Unknown Vendor ");
		break;
	}

	switch (info->flash_id & FLASH_TYPEMASK) {
	case FLASH_S29WS128P:
		printf ("FLASH S29WS128P\n");
		break;
	default:
		printf ("Unknown Chip Type\n");
		break;
	}

	printf ("  Size: %ld MB in %d Sectors\n",
			info->size >> 20, info->sector_count);

	printf ("  Sector Start Addresses:");
	for (i = 0; i < info->sector_count; ++i) {
		if ((i % 5) == 0)
			printf ("\n   ");
		printf (" %08lX%s",
			info->start[i], info->protect[i] ? " (RO)" : "     ");
	}
	printf ("\n");
	return;
}

/*
 * The following code cannot be run from FLASH!
 */
static ulong flash_get_size (FPW * addr, flash_info_t * info)
{
	volatile FPW value;
	FPWV *baddr = addr;

	/* Write auto select command: read Manufacturer ID */
	baddr[0x0] = (FPW) 0x00F0;
	baddr[0x0555] = (FPW) 0x00AA00AA;
	baddr[0x02AA] = (FPW) 0x00550055;
	baddr[0x0555] = (FPW) 0x00900090;

	mb ();
	value = baddr[0];

	switch (value) {
	case (FPW) AMD_MANUFACT:
		info->flash_id = FLASH_MAN_AMD;
		break;
	default:
		info->flash_id = FLASH_UNKNOWN;
		info->sector_count = 0;
		info->size = 0;
		baddr[0] = (FPW) 0x00F000F0;	/* restore read mode */
		return (0);						/* no or unknown flash	*/
	}

	mb ();
	value = baddr[1];	/* device ID	    */

	switch (value) {
    case (FPW) (AMD_ID_MIRROR):
		switch (baddr[0x0E]) {
		case (FPW) AMD_ID_WS128P_2:
			if (baddr[0x0F] == (FPW) AMD_ID_WS128P_3) {
				info->flash_id += FLASH_S29WS128P;
				info->sector_count = CFG_MAX_FLASH_SECT;
				info->size = PHYS_FLASH_SIZE;
			}
			else {
				info->flash_id = FLASH_UNKNOWN;
			}
			break;

		default:
			info->flash_id = FLASH_UNKNOWN;
			break;
		}
		break;
	default:
		info->flash_id = FLASH_UNKNOWN;
		break;
	}

	if (info->sector_count > CFG_MAX_FLASH_SECT) {
		printf ("** ERROR: sector count %d > max (%d) **\n",
				info->sector_count, CFG_MAX_FLASH_SECT);
		info->sector_count = CFG_MAX_FLASH_SECT;
	}

	baddr[0] = (FPW) 0x00F000F0; /* restore read mode */

	if (info->flash_id == FLASH_UNKNOWN)
		return (0);
	else
		return (info->size);
}


/* unprotects a sector for write and erase
 * on some intel parts, this unprotects the entire chip, but it
 * wont hurt to call this additional times per sector...
 */
void flash_unprotect_sectors (FPWV * addr)
{
	return;
}


/*-----------------------------------------------------------------------
 */

int flash_erase (flash_info_t * info, int s_first, int s_last)
{
	int flag, sect;
	ulong type, start, last;
	int rcode = ERR_OK;
	FPWV *baddr = (FPWV *) (info->start[0]);

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return ERR_INVAL;
	}

	type = (info->flash_id & FLASH_VENDMASK);
	if ((type != FLASH_MAN_AMD)) {
		printf ("Can't erase unknown flash type %08lx - aborted\n",
				info->flash_id);
		return ERR_UNKNOWN_FLASH_VENDOR;
	}

	printf ("\n");

	start = get_timer (0);
	last = start;

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts ();

	printf ("Erasing sector [%3d-%3d] ... %3d", s_first, s_last, s_first);

	/* Start erase on unprotected sectors */
	for (sect = s_first; sect <= s_last; sect++) {
		FPWV *addr = (FPWV *) (info->start[sect]);
		FPW status;

		printf ("\b\b\b%3d", sect);

		flash_unprotect_sectors (addr);

		/* arm simple, non interrupt dependent timer */
		reset_timer_masked ();

		baddr[0x0555] = (FPW) 0x00AA00AA;
		baddr[0x02AA] = (FPW) 0x00550055;
		baddr[0x0555] = (FPW) 0x00800080;
		baddr[0x0555] = (FPW) 0x00AA00AA;
		baddr[0x02AA] = (FPW) 0x00550055;
		addr[0] = (FPW) 0x00300030;

		/* wait at least 80us - let's wait 1 ms */
		udelay (1000);

		while (((status = *addr) & (FPW) 0x00800080) != (FPW) 0x00800080) {
			if (get_timer_masked () > CFG_FLASH_ERASE_TOUT) {
				printf ("Timeout\n");
				/* suspend erase     */
				*addr = (FPW) 0x00B000B0;
				/* reset to read mode */
				baddr[0] = (FPW) 0x00F000F0;
				rcode = ERR_TIMOUT;
				break;
			}
		}

		baddr[0] = (FPW) 0x00F000F0;
	}
	printf (" done\n");

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	return rcode;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 4 - Flash not identified
 */

int write_buff (flash_info_t * info, uchar * src, ulong addr, ulong cnt)
{
	ulong cp, wp;
	FPW data;
	int count, i, l, rc, port_width;
	ulong total;

	if (info->flash_id == FLASH_UNKNOWN) {
		return ERR_UNKNOWN_FLASH_VENDOR;
	}
/* get lower word aligned address */
#ifdef FLASH_PORT_WIDTH16
	wp = (addr & ~1);
	port_width = 2;
#else
	wp = (addr & ~3);
	port_width = 4;
#endif

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
		data = 0;
		for (i = 0, cp = wp; i < l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *) cp);
		}
		for (; i < port_width && cnt > 0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt == 0 && i < port_width; ++i, ++cp) {
			data = (data << 8) | (*(uchar *) cp);
		}

		if ((rc = write_data (info, wp, SWAP (data))) != 0) {
			return (rc);
		}
		wp += port_width;
	}

	/*
	 * handle word aligned part
	 */
	count = 0;
	total = cnt;
	spin_wheel (total, cnt);
	while (cnt >= port_width) {
		data = 0;
		for (i = 0; i < port_width; ++i) {
			data = (data << 8) | *src++;
		}
		if ((rc = write_data (info, wp, SWAP (data))) != 0) {
			return (rc);
		}
		wp += port_width;
		cnt -= port_width;
		if (count++ > 0x800) {
			spin_wheel (total, cnt);
			count = 0;
		}
	}

	spin_wheel (total, 0);
	if (cnt == 0) {
		return (0);
	}

	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i = 0, cp = wp; i < port_width && cnt > 0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i < port_width; ++i, ++cp) {
		data = (data << 8) | (*(uchar *) cp);
	}

	return (write_data (info, wp, SWAP (data)));
}

/*-----------------------------------------------------------------------
 * Write a word or halfword to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_data (flash_info_t * info, ulong dest, FPW data)
{
	FPWV *addr = (FPWV *) (info->start[0]);
	ulong status;
	int flag;

	/* Check if Flash is (sufficiently) erased */
	if ((*((FPWV *)dest) & data) != data) {
		printf ("not erased at %08lx (%x)\n", (ulong) addr, *addr);
		return (ERR_NOT_ERASED);
	}

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts ();

	addr[0x0555] = (FPW) 0x00AA00AA;
	addr[0x02AA] = (FPW) 0x00550055;
	addr[0x0555] = (FPW) 0x00A000A0;
	*((FPWV *)dest) = data;

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	/* arm simple, non interrupt dependent timer */
	reset_timer_masked ();

	/* wait while polling the status register */
	while ((status = *((FPWV *)dest)) != data) {
		if (get_timer_masked () > CFG_FLASH_WRITE_TOUT) {
			*addr = (FPW) 0x00F000F0;	/* restore read mode */
			return (ERR_TIMOUT);
		}
	}
	*addr = (FPW) 0x00F000F0;	/* restore read mode */

	return (ERR_OK);
}

void inline spin_wheel (ulong total, ulong cnt)
{
	if (total == cnt)
		printf ("[  0\%]");
	else
		printf ("\010\010\010\010\010\010[%3d\%]", (((total-cnt)*100)/total));
}

#endif	/* CONFIG_SONY_ICX */
