/*
 * i2c.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "../reg.h"
#include "icx_io.h"
#include "smu.h"
#include "i2c.h"


/****************/
/*@ Control I2C */
/****************/

/* IICC0 */
#define IIC_IICC0_SPT0				0x0001
#define IIC_IICC0_STT0				0x0002
#define IIC_IICC0_SPIE0				0x0010
#define IIC_IICC0_IICE0				0x0080

/* IICSE0 */
#define IIC_IICSE0_SPD0				0x0100
#define IIC_IICSE0_ACKD0			0x0400
#define IIC_IICSE0_MSTS0			0x8000

/* IICF0 */
#define IIC_IICF0_IICBSY			0x0040
#define IIC_IICF0_STCF				0x0080

/* INTC */
#define INTC_IT0_RAW_IIC2			0x00000080

/* timeout */
#define IIC_MAX_WAIT_STCF			1000
#define IIC_MAX_WAIT_SPD0			1000
#define IIC_MAX_WAIT_BSY			100
#define IIC_MAX_WAIT_SEND			1000


/***************/
/*@ PIN select */
/***************/

/* IIC2 */
#define	CHG_PINSEL_IIC2_MSK			0x00000003
#define	CHG_PINSEL_IIC2_IIC2		0x00000000

#define CHG_PULL0_MSK_IIC2			0x07000000
#define CHG_PULL0_IIC2				0x05000000

/* clock divisor */
#define	SMU_DIVIICSCLK_MSK_IIC2		0x00FF0000
#define	SMU_DIVIICSCLK_IIC2			0x00920000		/* (229.376MHz or 237.568MHz) / 40	*/


/*******************/
/*@ Local routines */
/*******************/

/*-----------------------------------------------------------------------------
 *  Function    : i2c_power_on
 *  Abstruct    : Power on I2C.
 *  Input		: none
 *  Return      : none
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
static void i2c_power_on(void)
{
	smu_clockgate_control(SMU_GCLKCTRL_IIC2_CLK,  SMU_GCLKCTRL_OPEN);
	smu_clockgate_control(SMU_GCLKCTRL_IIC2_SCLK, SMU_GCLKCTRL_OPEN);
	smu_reset_control(SMU_RESETREQ_IIC2, SMU_RESETREQ_UNRESET);
}

/*-----------------------------------------------------------------------------
 *  Function    : i2c_power_off
 *  Abstruct    : Power off I2C.
 *  Input		: none
 *  Return      : none
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
static void i2c_power_off(void)
{
	smu_clockgate_control(SMU_GCLKCTRL_IIC2_CLK,  SMU_GCLKCTRL_OPEN);
	smu_clockgate_control(SMU_GCLKCTRL_IIC2_SCLK, SMU_GCLKCTRL_OPEN);
	smu_reset_control(SMU_RESETREQ_IIC2, SMU_RESETREQ_RESET);
	smu_clockgate_control(SMU_GCLKCTRL_IIC2_SCLK, SMU_GCLKCTRL_CLOSE);
	smu_clockgate_control(SMU_GCLKCTRL_IIC2_CLK,  SMU_GCLKCTRL_CLOSE);
}

/*-----------------------------------------------------------------------------
 *  Function    : i2c_config
 *  Abstruct    : I2C configuration.
 *  Input       : config	: pointer to I2C config structure
 *  Return      : none
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
static void i2c_config(i2c_config_t *config)
{
	/* configuration */
	if (inh(IIC2_IICC0) & IIC_IICC0_IICE0) {
		outh(0, IIC2_IICC0);
		udelay(10000);
	}

	outh(config->clock, IIC2_IICCL0);
	outh(config->flag,  IIC2_IICF0);
	outh((config->ctrl | IIC_IICC0_IICE0 | IIC_IICC0_SPIE0), IIC2_IICC0);
}

/*-----------------------------------------------------------------------------
 *  Function    : i2c_wait_for_state
 *  Abstruct    : Wait for changing a status
 *  Input       : reg	: I2C register address
 *				: mask  : state bit
 *				: limit : timeout
 *				: cond  : 0 or 1
 *  Return      : 0 : changed a status / 1 : timeout
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
static int i2c_wait_for_state(unsigned int reg, unsigned short mask, int limit, int cond)
{
	int timeout = limit;

	/* waiting on Bus Busy */
	cond = cond ? 0 : mask;
	while (((inh(reg) & mask) == cond) && timeout--);

	if (timeout <= 0) 
		return 1;

	return 0;
}

#define	I2C_WAIT_FOR_BB()	 i2c_wait_for_state(IIC2_IICF0,  IIC_IICF0_IICBSY, IIC_MAX_WAIT_BSY,  0)
#define	I2C_WAIT_FOR_START() i2c_wait_for_state(IIC2_IICF0,  IIC_IICF0_STCF,   IIC_MAX_WAIT_STCF, 0)
#define	I2C_WAIT_FOR_STOP()  i2c_wait_for_state(IIC2_IICSE0, IIC_IICSE0_SPD0,  IIC_MAX_WAIT_SPD0, 1)

/*-----------------------------------------------------------------------------
 *  Function    : i2c_send_data
 *  Abstruct    : Send a data
 *  Input       : data		: data value
 *  Return      : 0 : success / -EIO, -ETIMEDOUT
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
static int i2c_send_data(unsigned char data)
{
	int timeout;

	/* clear the interrupt bit */
	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);

	/* write a data */
	outh(data, IIC2_IIC0);

	/* wait for interrupt status */
	for (timeout = IIC_MAX_WAIT_SEND; timeout > 0; timeout--) {
		if (inl(INTC_IT0_RAW1) & INTC_IT0_RAW_IIC2)
			break;
	}

	if (timeout <= 0)
		return -ETIMEDOUT;

	/* check ACK */
	if (!(inh(IIC2_IICSE0) & IIC_IICSE0_ACKD0))
		return -EIO;

	/* clear the interrupt bit */
	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : i2c_receive_data
 *  Abstruct    : receive a data
 *  Input       : data		: pointer to data 
 *  Return      : 0(success) / -EIO, -ETIMEDOUT 
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
static int i2c_receive_data(unsigned char *data, unsigned char isack)
{
	int timeout;
	unsigned short regval;

	/* clear the interrupt bit */
	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);

	/* receive setting */
	regval = inh(IIC2_IICC0);
	if (isack) {
		regval &= ~IIC_IICC0_WTIM0;
		regval |= IIC_IICC0_ACKE0;
	}
	else {
		regval |= IIC_IICC0_WTIM0;
		regval &= ~IIC_IICC0_ACKE0;
	}
	outh(regval, IIC2_IICC0);
	outh((regval | IIC_IICC0_WREL0), IIC2_IICC0);

	/* wait for interrupt status */
	for (timeout = IIC_MAX_WAIT_SEND; timeout > 0; timeout--) {
		if (inl(INTC_IT0_RAW1) & INTC_IT0_RAW_IIC2)
			break;
	}

	if (timeout <= 0)
		return -ETIMEDOUT;

	/* read a data */
	*data = inh(IIC2_IIC0);

	regval = inh(IIC2_IICC0);
	regval &= ~IIC_IICC0_ACKE0;
	outh(regval, IIC2_IICC0);

	/* clear the interrupt bit */
	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);

	return 0;
}


/********************/
/*@ Gloval routines */
/********************/

/*-----------------------------------------------------------------------------
 *  Function    : i2c_init
 *  Abstruct    : Initialize I2C.
 *  Input       : config	: pointer to I2C config structure
 *  Return      : 0:success / -EINVAL
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
int i2c_init(i2c_config_t *config) 
{
	unsigned int regval;

	if (config == NULL)
		return -EINVAL;

	/* pinsel */
	regval = inl(CHG_PINSEL_IIC2) & ~CHG_PINSEL_IIC2_MSK;
	outl(regval, CHG_PINSEL_IIC2);

	/* pull */
	regval = inl(CHG_PULL0) & ~CHG_PULL0_MSK_IIC2;
	outl((regval | CHG_PULL0_IIC2), CHG_PULL0);

	udelay(1000);

	/* set the divisor of system clock */
	regval = inl(ASMU_DIVIICSCLK) & ~SMU_DIVIICSCLK_MSK_IIC2;
	outl((regval | SMU_DIVIICSCLK_IIC2), ASMU_DIVIICSCLK);

	i2c_power_on();

	/* configuration */
	i2c_config(config);

	/* enable interrupt */
	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IEN1);

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : i2c_exit
 *  Abstruct    : Remove I2C.
 *  Input       : none
 *  Return      : none
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
int i2c_exit(void)
{
	/* disable interrupt */
	outl(INTC_IT0_IEN1, INTC_IT0_IDS1);

	/* stop I2C */
	if (inh(IIC2_IICC0) & IIC_IICC0_IICE0) {
		outh(0, IIC2_IICC0);
		udelay(10000);
	}

	i2c_power_off();

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : i2c_write
 *  Abstruct    : Write a data to I2C.
 *  Input       : config 	: pointer to I2C config structure
 *				: reg		: register address
 *				: data		: data value
 *  Return      : 0(success) / -EIO, -ETIMEDOUT 
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
int i2c_write(i2c_config_t *config, unsigned char reg, unsigned char data)
{
	int ret = 0;

	/* waiting on Bus Busy */
	if (I2C_WAIT_FOR_BB()) {
		if (!(inh(IIC2_IICSE0) & IIC_IICSE0_MSTS0))
			return -ETIMEDOUT;

		outh((inh(IIC2_IICC0) | IIC_IICC0_SPT0), IIC2_IICC0);

		if (I2C_WAIT_FOR_BB()) {
			i2c_config(config);
			if (I2C_WAIT_FOR_BB())
				return -ETIMEDOUT;
		}
	}

	/* start condition */
	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);
	outh((inh(IIC2_IICC0) | IIC_IICC0_STT0), IIC2_IICC0);
	if (I2C_WAIT_FOR_START())
		return -ETIMEDOUT;

	/* send slave address */
	if ((ret = i2c_send_data(config->slave_adr)) == 0) {
		/* send register address */
		if ((ret = i2c_send_data(reg)) == 0) {
			/* send a data */
			ret = i2c_send_data(data);
		}
	}

	/* stop condition */
	outh((inh(IIC2_IICC0) | IIC_IICC0_SPT0), IIC2_IICC0);
	if (I2C_WAIT_FOR_STOP())
		return -ETIMEDOUT;

	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);

	return ret;
}

/*-----------------------------------------------------------------------------
 *  Function    : i2c_read
 *  Abstruct    : Read a data from I2C.
 *  Input       : config 	: pointer to I2C config structure
 *				: reg		: register address
 *				: data		: pointer to the data
 *  Return      : 0(success) / -EIO, -ETIMEDOUT 
 *  Note		: use only IIC2
 *----------------------------------------------------------------------------*/
int i2c_read(i2c_config_t *config, unsigned char reg, unsigned char *data, int num)
{
	int i;
	int ret = 0;

	/* waiting on Bus Busy */
	if (I2C_WAIT_FOR_BB()) {
		if (!(inh(IIC2_IICSE0) & IIC_IICSE0_MSTS0))
			return -ETIMEDOUT;

		outh((inh(IIC2_IICC0) | IIC_IICC0_SPT0), IIC2_IICC0);

		if (I2C_WAIT_FOR_BB()) {
			i2c_config(config);
			if (I2C_WAIT_FOR_BB())
				return -ETIMEDOUT;
		}
	}

	/* start condition */
	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);
	outh((inh(IIC2_IICC0) | IIC_IICC0_STT0), IIC2_IICC0);
	if (I2C_WAIT_FOR_START())
		return -ETIMEDOUT;

	/* send slave address */
	if ((ret = i2c_send_data(config->slave_adr)) == 0) {
		/* send register address */
		if ((ret = i2c_send_data(reg)) == 0) {

			/* start condition */
			outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);
			outh((inh(IIC2_IICC0) | IIC_IICC0_STT0), IIC2_IICC0);
			if (I2C_WAIT_FOR_START()) {
				ret = -ETIMEDOUT;
			}
			else {
				/* send slave address with read flag */
				if ((ret = i2c_send_data(config->slave_adr|0x01)) == 0) {
					/* receive a data */
					for (i = 0; i < num; i++) {
						if (i == (num-1))
							ret = i2c_receive_data(&data[i], 0);
						else
							ret = i2c_receive_data(&data[i], 1);
					}
				}
			}
		}
	}

	/* stop condition */
	outh((inh(IIC2_IICC0) | IIC_IICC0_SPT0), IIC2_IICC0);
	if (I2C_WAIT_FOR_STOP())
		return -ETIMEDOUT;

	outl(INTC_IT0_RAW_IIC2, INTC_IT0_IIR);

	return ret;
}

#endif	/* CONFIG_ICX */

