/*
 * icx.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include "icx.h"
#include "md5.h"

#include "pwc.h"
#include "spi.h"
#include "adc.h"
#include "lcdc.h"

#include "icx_pm.h"
#include "icx_key.h"
#include "icx_lcdc.h"
#include "icx_beep.h"
#include "icx_emmc.h"
#include "icx_nvp.h"

#include "../../../../../kernel/top/include/asm-arm/arch-mp200/icx_emmc_map.h"

/******************/
/* from emmc-mini */
/******************/

#include "../emmc_mini/emmc/emmc.h"

/* boot information */
struct boot_info {
	unsigned long addr;		/* ddr address	*/
	unsigned long block;	/* block number	*/
	unsigned long size;		/* size [byte]	*/
	unsigned long erase;	/* erase flag	*/
};

/* prototype */
extern void emmc_init(void);
extern int  emmc_read(struct boot_info *boot_info);


/***********************/
/*@ Initialize routine */
/***********************/

/* boot informations */
unsigned int   icx_pm_boot_factor = 0;
unsigned char  icx_pm_psw_state	  = 0;
unsigned char  icx_pm_vbat  	  = 0;
unsigned char  icx_pm_vbat_low 	  = 0;
unsigned char  icx_pm_vbat_vlow	  = 0;
icx_pm_date_t  icx_pm_date		  = {0};
unsigned short icx_key_value	  = ICX_KEY_NOT_PRESSED;

/* boot action flag */
typedef struct {
	int	power_switch;
	int	lcdc_switch;
	int	beep_switch;
} icx_action_flag_t;

/*-----------------------------------------------------------------------------
 *  Function    : icx_init_sequence_adaptor
 *  Abstruct    : AC adaptor boot
 *  Input       : action_flag
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void icx_init_sequence_adaptor(icx_action_flag_t *action_flag)
{
	icx_pm_set_vbus_limit(100);

	/* BEEP switch ON */
	action_flag->beep_switch = ICX_BEEP_ON;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_select_clock_mode
 *  Abstruct    : select clk mode
 *  Input       : normal_div
 *  Return      : none
 *----------------------------------------------------------------------------*/
/* clock mode */
#define ICX_CLOCK_MODE_NORMAL_A	1
#define ICX_CLOCK_MODE_NORMAL_B	2
#define ICX_CLOCK_MODE_NORMAL_C	3
#define ICX_CLOCK_MODE_NORMAL_D	4

static void icx_select_clock_mode(unsigned int mode)
{
	ulong auto_frq_change;

	if ((inl(ASMU_CLK_MODE_SEL) & 0x00000F00) == (mode << 8)) {
		return;
	}

	/* auto freq change off */
	auto_frq_change = inl(ASMU_AUTO_FRQ_CHANGE);
	outl((auto_frq_change & ~0x1), ASMU_AUTO_FRQ_CHANGE);

	outl(mode, ASMU_CLK_MODE_SEL);

	while ((inl(ASMU_CLK_MODE_SEL) & 0x00000F00) != (mode << 8));

	/* restore auto freq change */
	outl(auto_frq_change, ASMU_AUTO_FRQ_CHANGE);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_init_sequence_usb
 *  Abstruct    : USB boot
 *  Input       : action_flag
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void icx_init_sequence_usb(icx_action_flag_t *action_flag)
{
	if (icx_pm_check_battery_flag()) {
		/* battery is not connected */
		icx_pm_save_status_sram(ICX_PM_BATT_STATUS_VLOW | ICX_PM_BATT_STATUS_NONE);
		action_flag->lcdc_switch = ICX_LCDC_OFF;

		/* set clock mode to NORMAL_D */
		icx_select_clock_mode(ICX_CLOCK_MODE_NORMAL_D);
	}
	else if (icx_pm_vbat < icx_pm_vbat_vlow) {
		/* VERY_LOW_BATT	*/
		icx_pm_save_status_sram(ICX_PM_BATT_STATUS_VLOW);
		action_flag->lcdc_switch = ICX_LCDC_OFF;

		/* set clock mode to NORMAL_D */
		icx_select_clock_mode(ICX_CLOCK_MODE_NORMAL_D);
	}
	else {
		icx_pm_set_vbus_limit(100);

		/* set battery connected flag	*/
		icx_pm_set_battery_flag(ICX_PM_BATT_FLAG_SET);
	}

	/* BEEP switch ON */
	action_flag->beep_switch = ICX_BEEP_ON;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_init_sequence_rtc
 *  Abstruct    : RTC alarm boot
 *  Input       : action_flag
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void icx_init_sequence_rtc(icx_action_flag_t *action_flag)
{
	if (icx_pm_vbat < icx_pm_vbat_vlow) {
		/* VeryLowBatt */
		action_flag->power_switch = ICX_PM_OFF;
	}
	else {
		icx_pm_set_vbus_limit(100);

		if (icx_pm_vbat < icx_pm_vbat_low) {
			/* LowBatt */
			action_flag->power_switch = ICX_PM_OFF;
		}
		else {
			if (icx_pm_check_rtc_date(&icx_pm_date)) {
				icx_pm_save_status_sram(ICX_PM_BATT_STATUS_RTC);

				action_flag->lcdc_switch = ICX_LCDC_OFF;
			}
			else {
				action_flag->power_switch = ICX_PM_OFF;
			}
		}
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_init_sequence_cradle
 *  Abstruct    : Cradle boot
 *  Input       : action_flag
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void icx_init_sequence_cradle(icx_action_flag_t *action_flag)
{
	if (icx_pm_vbat < icx_pm_vbat_vlow) {
		/* VeryLowBatt */
		action_flag->power_switch = ICX_PM_OFF;
	}
	else {
		icx_pm_set_vbus_limit(100);

		if (icx_pm_vbat < icx_pm_vbat_low) {
			/* LowBatt */

			/* check to connect ADP/USB */
			if (icx_pm_check_cancel_vbat_low() == ICX_PM_LOWBATT_CHK_TIMEOUT) {
				action_flag->power_switch = ICX_PM_OFF;
			}
		}
		else {
			/* BEEP switch ON */
			action_flag->beep_switch = ICX_BEEP_ON;
		}
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_init_sequence_key
 *  Abstruct    : KEY wake up
 *  Input       : action_flag
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void icx_init_sequence_key(icx_action_flag_t *action_flag)
{
	unsigned char hold_state;
	int ret;

	if (icx_pm_vbat < icx_pm_vbat_vlow) {
		/* VeryLowBatt */
		action_flag->power_switch = ICX_PM_OFF;
	}
	else {
		icx_pm_set_vbus_limit(100);

		if (icx_pm_vbat < icx_pm_vbat_low) {
			/* LowBatt */

			/* check to connect ADP/USB */
			if (icx_pm_check_cancel_vbat_low() == ICX_PM_LOWBATT_CHK_TIMEOUT) {
				action_flag->power_switch = ICX_PM_OFF;
			}
		}
		else {
			icx_key_get_hold(&hold_state);

			if ((hold_state == ICX_KEY_HOLD_ON)
			 &&	(icx_nvp_check_hold_mode() == ICX_NVP_HOLD_MODE_ALL)) {
				/* HOLD ON */

				/* check to cancel HOLD */
				ret = icx_key_check_cancel_hold();

				/* set default KEY */
				icx_key_value = ICX_KEY_NOT_PRESSED;

				if (ret == ICX_KEY_HOLD_CHK_TIMEOUT) {
					action_flag->power_switch = ICX_PM_OFF;
				}
				else if (ret == ICX_KEY_HOLD_CHK_CANCEL_RTC) {
					/* power off LCD */
					icx_lcdc_power_off();
					action_flag->lcdc_switch = ICX_LCDC_OFF;

					icx_pm_save_status_sram(ICX_PM_BATT_STATUS_RTC);
				}
			}
			else {
				/* BEEP switch ON */
				action_flag->beep_switch = ICX_BEEP_ON;
			}
		}
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_init_sequence_hold
 *  Abstruct    : HOLD wake up
 *  Input       : action_flag
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void icx_init_sequence_hold(icx_action_flag_t *action_flag)
{
	unsigned char hold_state;

	if (icx_pm_vbat < icx_pm_vbat_vlow) {
		/* VeryLowBatt */
		action_flag->power_switch = ICX_PM_OFF;
	}
	else {
		icx_key_get_hold(&hold_state);
		if (hold_state == ICX_KEY_HOLD_ON) {
			/* HOLD OFF -> ON */
			action_flag->power_switch = ICX_PM_OFF;
		}
		else {
			/* HOLD ON -> OFF */

			icx_pm_set_vbus_limit(100);

			if (icx_pm_vbat < icx_pm_vbat_low) {
				/* LowBatt */

				/* check to connect ADP/USB */
				if (icx_pm_check_cancel_vbat_low() == ICX_PM_LOWBATT_CHK_TIMEOUT) {
					action_flag->power_switch = ICX_PM_OFF;
				}
			}
		}
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_init_sequence_others
 *  Abstruct    : Battery boot
 *  Input       : action_flag
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void icx_init_sequence_others(icx_action_flag_t *action_flag)
{
	if (icx_pm_vbat < icx_pm_vbat_vlow) {
		/* VeryLowBatt */
		action_flag->power_switch = ICX_PM_OFF;
	}
	else {
		icx_pm_set_vbus_limit(100);

		if (icx_pm_vbat < icx_pm_vbat_low) {
			/* LowBatt */

			/* check to connect ADP/USB */
			if (icx_pm_check_cancel_vbat_low() == ICX_PM_LOWBATT_CHK_TIMEOUT) {
				action_flag->power_switch = ICX_PM_OFF;
			}
		}
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_init
 *  Abstruct    : Initialize ICX 
 *  Input       : none
 *  Return      : 0(success)
 *----------------------------------------------------------------------------*/
int icx_init(void)
{
	icx_action_flag_t action_flag = {ICX_PM_ON, ICX_LCDC_ON, ICX_BEEP_OFF};

	/***************************/
	/* Initialize EM1's macros */
	/***************************/
	spi_init();
	pwc_init();
	adc_init();
	emmc_init();
	lcdc_init();

	icx_emmc_init();

	/********************/
	/* Get boot factors */
	/********************/
	icx_key_value	   = icx_key_get_key_value();
	icx_pm_get_rtc_date(&icx_pm_date);
	icx_pm_vbat		   = icx_pm_get_vbat();
	icx_pm_boot_factor = icx_pm_get_boot_factor();
	icx_pm_psw_state   = icx_pm_get_psw_state();

	icx_pm_get_vbat_threshold(&icx_pm_vbat_low, &icx_pm_vbat_vlow);

	/**********************/
	/* Initialize drivers */
	/**********************/
	icx_nvp_init();
	icx_pm_init();
	icx_key_init();
	icx_lcdc_init();
	icx_beep_init();

	/**************************/
	/* Select a boot sequence */
	/**************************/

	/* skip this function when install */
	if( !ICX_BOOT_MODE_CHK(ICX_BOOT_MODE_MS)
	&&  !ICX_BOOT_MODE_CHK(ICX_BOOT_MODE_SD) ){
		if (icx_pm_psw_state & PWC_PSW_STATE_SDETADP) {
			/* AC adaptor	*/
			icx_init_sequence_adaptor(&action_flag);
		}
		else if (icx_pm_psw_state & PWC_PSW_STATE_SDETUSB) {
			/* USB			*/
			icx_init_sequence_usb(&action_flag);
		}
		else if (icx_pm_boot_factor & ICX_PM_FACTOR_RTC) {
			/* RTC alarm	*/
			icx_init_sequence_rtc(&action_flag);
		}
		else if ((icx_pm_boot_factor & ICX_PM_FACTOR_KEY) || (icx_key_value != ICX_KEY_NOT_PRESSED)) {
			/* KEY wakeup	*/
			icx_init_sequence_key(&action_flag);
		}
		else if (icx_pm_boot_factor & ICX_PM_FACTOR_HOLD) {
			/* HOLD wakeup	*/
			icx_init_sequence_hold(&action_flag);
		}
		else if ((icx_pm_boot_factor & ICX_PM_FACTOR_CRADLE) && (icx_key_value == ICX_KEY_NOT_PRESSED)) {
			/* CRADLE		*/
			icx_init_sequence_cradle(&action_flag);
		}
		else {
			/* others		*/
			icx_init_sequence_others(&action_flag);
		}
	}

	/************************/
	/* Execute boot actions */
	/************************/

	if (action_flag.power_switch == ICX_PM_ON) {
		/* Power ON */

#if 0
		if ((action_flag.beep_switch == ICX_BEEP_ON)
		 && (icx_key_check_key_code(icx_key_value, ICX_KEY_CODE_PLAY))
		 && (icx_nvp_check_beep_flag())) {
			/* Beep ON */
			icx_beep_enable();
			icx_beep_on(ICX_BEEP_TIME_100MS);
		}
#endif
		icx_beep_disable();

		icx_lcdc_draw_image(ICX_NVP_ID_BOOT_IMAGE);
		if (action_flag.lcdc_switch == ICX_LCDC_ON) {
			/* LCD ON */
			icx_lcdc_power_on();
			icx_lcdc_switch(ICX_LCDC_ON);
			icx_lcdc_backlight_control(ICX_LCDC_BRIGHTNESS_MAX, ICX_LCDC_FADE_INTERVAL_OPENING);
		}

		icx_key_save_key_value(icx_key_value);
	}
	else {
		/* Power OFF */

		/* clear battery connected flag	*/
		icx_pm_set_battery_flag(ICX_PM_BATT_FLAG_CLR);

		icx_lcdc_power_off();
		icx_beep_exit();
		icx_emmc_exit();

		icx_pm_power_off();
	}

	return 0;
}


/*****************/
/*@ Boot routine */
/*****************/

/* Partition */
#define	ICX_FULINUX_SECTOR		(LOADER_AREA_SECTORS)
#define	ICX_KERNEL_SECTOR		(LOADER_AREA_SECTORS + FU_LINUX_AREA_SECTORS + NVP_AREA_SECTORS)

#define ICX_LINUX_SIZE_OFFSET	4
#define ICX_LINUX_SUM_SIZE		4

/* Bootargs */
#ifdef CONFIG_SONY_ICX
#define	ICX_BOOTARGS_INST		"root=/dev/ram init=/linuxrc_inst SELINUX_INIT=no mem=32M rw video=qfb: ip=none"
#define	ICX_BOOTARGS_FWUP		"root=/dev/ram init=/linuxrc_fwup SELINUX_INIT=no mem=32M rw video=qfb: ip=none"
#define	ICX_BOOTARGS_EMRG		"root=/dev/ram init=/linuxrc SELINUX_INIT=no mem=32M rw video=qfb: ip=none"
#else
#define	ICX_BOOTARGS_INST		"root=/dev/ram init=/linuxrc_inst SELINUX_INIT=no mem=32M rw video=qfb: ip=none"
#define	ICX_BOOTARGS_FWUP		"root=/dev/ram init=/linuxrc_fwup SELINUX_INIT=no mem=32M rw video=qfb: ip=none"
#define	ICX_BOOTARGS_EMRG		"root=/dev/ram init=/linuxrc SELINUX_INIT=no mem=32M rw video=qfb: ip=none"
#endif

/* Linux load address */
#define	ICX_KERNEL_ADDR			0x30008000
#define	ICX_INITRD_ADDR			(TEXT_BASE + LOADER_AREA_SECTORS * SDM_BLOCKLEN_VAL)

/* misc */
#define	ICX_REVISE_BOUNDARY(s, b)	((s + (b - 1)) & ~(b - 1))


/*-----------------------------------------------------------------------------
 *  Function    : Boot error
 *  Abstruct    :
 *  Input       : none
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void boot_error(void)
{
	printf("!! Failed to boot Linux!\n");
	while(1);
}

/*-----------------------------------------------------------------------------
 *  Function    : load_linux_image
 *  Abstruct    :
 *  Input       : unsigned int   src       : sector number
 *              : unsigned char *dst       : load address
 *              : unsigned int  size_limit : limit of size
 *  Return      : readed size / 0(NG)
 *----------------------------------------------------------------------------*/
static int load_linux_image(unsigned int src, unsigned char *dst, unsigned int size_limit)
{
	struct boot_info emmcinfo;
	unsigned char emmc_buf[SDM_BLOCKLEN_VAL];
	unsigned int  size, size_total, size_1st;
	unsigned int *bufp;
	unsigned int  sum;
	int i;
	int ret;

	/******************/
	/* get image size */
	/******************/

	/* set parameter */
	emmcinfo.addr  = (unsigned long)emmc_buf;
	emmcinfo.block = src;
	emmcinfo.size  = SDM_BLOCKLEN_VAL;

	/* load 1st sector */
	ret = emmc_read(&emmcinfo);
	if (ret) {
		printf("Failed to read(eMMC info: %08Xh).\n", ret);
		return 0;
	}

	size_1st = SDM_BLOCKLEN_VAL - ICX_LINUX_SIZE_OFFSET;
	memcpy(dst, &emmc_buf[ICX_LINUX_SIZE_OFFSET], size_1st);

	size = inl(emmc_buf);
	size_total = size + ICX_LINUX_SIZE_OFFSET + ICX_LINUX_SUM_SIZE;
	size_total = ICX_REVISE_BOUNDARY(size_total, SDM_BLOCKLEN_VAL);

	if (size_total > size_limit) {
		printf("Too large image(%08Xh) - remain:%08Xh.\n", size_total, size_limit);
		return 0;
	}	

	/****************/
	/* read remains */
	/****************/

	/* set parameter */
	emmcinfo.addr  = (unsigned long)dst + size_1st;
	emmcinfo.block = src + 1;
	emmcinfo.size  = size_total - SDM_BLOCKLEN_VAL;

	ret = emmc_read(&emmcinfo);
	if (ret) {
		printf("Failed to read(eMMC info: %08Xh).\n", ret);
		return 0;
	}

	/************/
	/* calc sum */
	/************/

	sum  = 0;
	bufp = (unsigned int *)dst;
	size_total = (size_total - ICX_LINUX_SIZE_OFFSET) / 4;
	for (i = 0; i < size_total; i++) {
		sum ^= *bufp++;
	}
	if (sum) {
		printf("Sum error(%08Xh).\n", sum);
		return 0;
	}

	return size;	/* succeeded to load */
}

/*-----------------------------------------------------------------------------
 *  Function    : setup_tag
 *  Abstruct    :
 *  Input       : bd_t		  *bd			: U-Boot environment variables
 *              : char 		  *commandline	: kernel command line
 *              : unsigned int initrd_size	: size of InitRD
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void setup_tag(bd_t *bd, char *commandline, unsigned int initrd_size)
{
	struct tag *params;
	unsigned int sim;
	char *p;

	params = (struct tag *)bd->bi_boot_params;

	/********/
	/* core */
	/********/
	params->hdr.tag  = ATAG_CORE;
	params->hdr.size = tag_size(tag_core);

	params->u.core.flags    = 0;
	params->u.core.pagesize = 0;
	params->u.core.rootdev  = 0;

	params = tag_next(params);

	/**********/
	/* memory */
	/**********/

	sim=icx_nvp_get_sysinfo_mem();

	if(sim==ICX_NVP_SYSINFO_MEM_128M){
		params->hdr.tag = ATAG_MEM;
		params->hdr.size = tag_size (tag_mem32);
		params->u.mem.start = 0x30000000;
		params->u.mem.size =  0x02000000;
		params = tag_next (params);

		params->hdr.tag = ATAG_MEM;
		params->hdr.size = tag_size (tag_mem32);
		params->u.mem.start = 0x34000000;
		params->u.mem.size =  0x02000000;
		params = tag_next (params);

		params->hdr.tag = ATAG_MEM;
		params->hdr.size = tag_size (tag_mem32);
		params->u.mem.start = 0x36000000;
		params->u.mem.size =  0x02000000;
		params = tag_next (params);
	}
	else{ /* 64M */
		params->hdr.tag = ATAG_MEM;
		params->hdr.size = tag_size (tag_mem32);
		params->u.mem.start = 0x30000000;
		params->u.mem.size =  0x02000000;
		params = tag_next (params);

		params->hdr.tag = ATAG_MEM;
		params->hdr.size = tag_size (tag_mem32);
		params->u.mem.start = 0x32000000;
		params->u.mem.size =  0x00D00000;
		params = tag_next (params);
	}

	/****************/
	/* command line */
	/****************/
	if (commandline) {
		for (p = commandline; *p == ' '; p++);
		if (*p != '\0') {
			params->hdr.tag  = ATAG_CMDLINE;
			params->hdr.size = (sizeof (struct tag_header) + strlen (p) + 1 + 4) >> 2;

			strncpy(params->u.cmdline.cmdline, p, strlen(p) + 1);

			params = tag_next(params);
		}
	}

	/**********/
	/* initrd */
	/**********/
	if (initrd_size) {
		params->hdr.tag  = ATAG_INITRD2;
		params->hdr.size = tag_size(tag_initrd);

		params->u.initrd.start = ICX_INITRD_ADDR;
		params->u.initrd.size  = initrd_size;

		params = tag_next(params);
	}

	/*********************************/
    /* revision(=system information) */
	/*********************************/
	params->hdr.tag  = ATAG_REVISION;
	params->hdr.size = tag_size(tag_revision);

	params->u.revision.rev = icx_nvp_get_sysinfo();

	params = tag_next(params);

	/***********/
	/* end tag */
	/***********/
	params->hdr.tag  = ATAG_NONE;
	params->hdr.size = 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : emmc_boot_linux
 *  Abstruct    :
 *  Input       : unsigned int boot_mode : (Normal / Emergency F/W update)
 *  Return      : none
 *----------------------------------------------------------------------------*/
void emmc_boot_linux(unsigned int boot_mode)
{
	DECLARE_GLOBAL_DATA_PTR;

	bd_t *bd = gd->bd;
	char *commandline;
	int	kernel_sector;
	unsigned int	kernel_size;
	unsigned int	initrd_sector, initrd_size;
	int				boot_mode_flag = ICX_BOOT_FWUP;
	void (*kernel_entry)(int zero, int arch, uint params);
	
	/***************************/
	/* set kernel command line */
	/***************************/

	/* check the boot mode and change the kernel command line */
	if (ICX_BOOT_MODE_CHK(ICX_BOOT_MODE_MS))
		setenv("bootargs", ICX_BOOTARGS_INST);				/* installation	*/
	else if(ICX_BOOT_MODE_CHK(ICX_BOOT_MODE_SD))
		setenv("bootargs", ICX_BOOTARGS_INST);				/* installation	*/
	else if (boot_mode == ICX_BOOT_EMERGENCY)
		setenv("bootargs", ICX_BOOTARGS_EMRG);				/* emergency	*/
	else if (icx_nvp_check_fw_update())
		setenv("bootargs", ICX_BOOTARGS_FWUP);				/* F/W update	*/
	else
		boot_mode_flag = ICX_BOOT_NORMAL;

	/*********************/
	/* get kernel sector */
	/*********************/

	kernel_sector = icx_emmc_get_first_sector();
	if (kernel_sector < 0) {
		printf("\nFailed to read MBR.\n");
		boot_error();		/* never return */
	}

	if (boot_mode_flag == ICX_BOOT_NORMAL) {
		printf("\nLoading Linux...\n");
		kernel_size    = COMMON_KERNEL_AREA_SECTORS * SDM_BLOCKLEN_VAL;
		kernel_sector += ICX_KERNEL_SECTOR;
		initrd_size = 0;
	}
	else {
		printf("\nLoading Linux(for F/W update)...\n");
		kernel_size    = FU_LINUX_AREA_SECTORS * SDM_BLOCKLEN_VAL;
		kernel_sector += ICX_FULINUX_SECTOR;
		initrd_size = 1;
	}

	/***************/
	/* load kernel */
	/***************/

	kernel_size = load_linux_image(kernel_sector, (unsigned char *)ICX_KERNEL_ADDR, kernel_size);
	if (kernel_size == 0) {
		printf("Failed to load kernel image.\n");
		boot_error();		/* never return */
	}

	/********************************/
	/* load initrd(only F/W update) */
	/********************************/

	if (initrd_size) {
		/* get starting sector */
		kernel_size += ICX_LINUX_SIZE_OFFSET + ICX_LINUX_SUM_SIZE;
		kernel_size  = ICX_REVISE_BOUNDARY(kernel_size, SDM_BLOCKLEN_VAL);
		initrd_sector = kernel_sector + (kernel_size / SDM_BLOCKLEN_VAL);

		initrd_size = load_linux_image(initrd_sector,
									   (unsigned char *)ICX_INITRD_ADDR,
									   ((FU_LINUX_AREA_SECTORS*SDM_BLOCKLEN_VAL) - kernel_size));
		if (initrd_size == 0) {
			printf("Failed to load initrd image.\n");
			boot_error();	/* never return */
		}
	}

	/***************/
	/* boot Kernel */
	/***************/

	/* set entry point */
	kernel_entry = (void (*)(int, int, uint))ICX_KERNEL_ADDR;

	/* create TAG */
	commandline = getenv("bootargs");
	setup_tag(bd, commandline, initrd_size);

	cleanup_before_linux();

	printf("\nStarting kernel ...\n\n");

	if (boot_mode_flag == ICX_BOOT_FWUP) {
		icx_pm_set_battery_flag(ICX_PM_BATT_FLAG_CLR);
	}

	/* BOOT_DEBUG_LED */
	/* *(volatile unsigned int *)0xC0050000=0x80000000; */
	/* *(volatile unsigned int *)0xC0050040=0x00000001; */
	/* *(volatile unsigned int *)0xC005000C=0x80000000; */
	/* *(volatile unsigned int *)0xC0050048=0x00010001; */

	/* jump to kernel */
	kernel_entry(0, bd->bi_arch_number, bd->bi_boot_params);
}


/*********************/
/*@ Password routine */
/*********************/

/*-----------------------------------------------------------------------------
 *  Function    : chr2asc
 *  Abstruct    : 
 *  Input       : char *dst : buffer address to save ASCII code 
 *              : char  src : char code
 *  Return      : none
 *----------------------------------------------------------------------------*/
static void chr2asc(char *dst, char src)
{
	int i;
	char c;

	for (i = 1; i >= 0; i--) {
		c = ((unsigned int)src >> (4 * i)) & 0xF;
		if ((unsigned int)c < 10)
			c = c + '0';
		else
			c = c - 10 + 'a';
		dst[(1-i)] = c;
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : check_pwd
 *  Abstruct    : 
 *  Input       : char *pwd : address of password
 *              : int   len : length of password
 *  Return      : 0(OK) / 1(NG)
 *----------------------------------------------------------------------------*/
int check_pwd(const char *pwd, int len)
{
	char		  pwdbuf[ICX_BOOT_PWD_LEN], buf[ICX_BOOT_PWD_LEN];
	MD5_CTX		  md5c;
	unsigned char sig[16];
	int           i;
	int           rc;

	/* get pwd from NVP */
	icx_nvp_get_pwd(pwdbuf);

	if (len > ICX_BOOT_PWD_LEN)
		return 0;

	/* calcurate MD5 digest */
	MD5Init(&md5c);
	MD5Update(&md5c, (unsigned char *)pwd, len);
	MD5Final(sig, &md5c);

	for (i = 0; i < 16; i++) 
		chr2asc(&buf[i*2], sig[i]);

	/* compare pwds */
	if (strncmp(pwdbuf, buf, ICX_BOOT_PWD_LEN) == 0)
		rc = 1;
	else
		rc = 0;

	return rc;
}

#endif /* CONFIG_ICX */

