/*
 * icx_key.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "icx_io.h"
#include "pwc.h"
#include "adc.h"
#include "icx_pm.h"
#include "icx_key.h"
#include "icx_nvp.h"
#include "icx_lcdc.h"


/****************/
/*@ Definitions */
/****************/

/* SRAM address(to save key A/D value) */
#define	ICX_KEY_SRAM_ADR		0xA0000104
#define	ICX_KEY_SRAM_MAGIC		0x794B0000			/* "Ky"	*/

#define ICX_KEY_MASK_ALL		0x0000FFFF

/* GPIO */
#define	ICX_KEY_GPIO_WAKE		PWC_GPIO_IO00
#define	ICX_KEY_GPIO_XHOLD		PWC_GPIO_IO01
#define	ICX_KEY_GPIO_NC_ON		PWC_GPIO_IO02

#define ICX_KEY_GPPUPD_WAKE			(PWC_GPIO_GPPUPD_OPEN << PWC_GPIO_GPPUPD_SFT_IO00)
#define ICX_KEY_GPPUPD_WAKE_MSK		(PWC_GPIO_GPPUPD_MSK  << PWC_GPIO_GPPUPD_SFT_IO00)
#define ICX_KEY_GPPUPD_XHOLD		(PWC_GPIO_GPPUPD_OPEN << PWC_GPIO_GPPUPD_SFT_IO01)
#define ICX_KEY_GPPUPD_XHOLD_MSK	(PWC_GPIO_GPPUPD_MSK  << PWC_GPIO_GPPUPD_SFT_IO01)
#define ICX_KEY_GPPUPD_NC_ON		(PWC_GPIO_GPPUPD_OPEN << PWC_GPIO_GPPUPD_SFT_IO02)
#define ICX_KEY_GPPUPD_NC_ON_MSK	(PWC_GPIO_GPPUPD_MSK  << PWC_GPIO_GPPUPD_SFT_IO02)

/* A/D ch */
#define	ICX_KEY_ADSEL_KEY0		PWC_ADSEL_ADIN3
#define	ICX_KEY_ADSEL_KEY1		PWC_ADSEL_ADIN4

/* key value */
#define ICX_KEY_VALUE_DEFAULT	0xFF				/* not pressed  */

#define ICX_KEY_KEY1_SFT		8

#define ICX_KEY_NUM				2

/* HOLD */
#define	ICX_KEY_HOLD_CHAT_INTERVAL	30000			/* 30ms			*/

#define	ICX_KEY_HOLD_CHK_INTERVAL	10000			/* 10ms			*/
#define	ICX_KEY_HOLD_CHK_COUNT		500				/* 5s			*/


/********************/
/*@ Local variables */
/********************/

/* key code table */
const static unsigned char icx_key_value_table[] = {
	29, 61, 93, 125, 158, 192, 224, 247
};


/********************/
/*@ Global routines */
/********************/

/*-----------------------------------------------------------------------------
 *  Function    : icx_key_init
 *  Abstruct    : Initialize KEY driver
 *  Input       : none
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_key_init(void)
{
	/* set GPIO input */
	pwc_write(PWC_GPIO_IOSEL1, PWC_GPIO_IOSEL_IN, (ICX_KEY_GPIO_WAKE 
												  |ICX_KEY_GPIO_XHOLD
												  |ICX_KEY_GPIO_NC_ON));

	/* set pull */
	pwc_write(PWC_GPIO_GPPUPD1,
			  (ICX_KEY_GPPUPD_WAKE
			  |ICX_KEY_GPPUPD_XHOLD
			  |ICX_KEY_GPPUPD_NC_ON),
			  (ICX_KEY_GPPUPD_WAKE_MSK
			  |ICX_KEY_GPPUPD_XHOLD_MSK
			  |ICX_KEY_GPPUPD_NC_ON_MSK));
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_key_get_key_value
 *  Abstruct    : Get key A/D value
 *  Input       : none
 *  Return      : key code
 *				:	bit15-8: KEY1
 *				:	bit 7-0: KEY0
 *----------------------------------------------------------------------------*/
unsigned short icx_key_get_key_value(void)
{
	unsigned short key_value = 0;
	unsigned char  key_ad;
	int i;

	/*********************/
	/* get key A/D value */
	/*********************/
	for (i = 0; i < ICX_KEY_NUM; i++) {
		if (adc_read((ICX_KEY_ADSEL_KEY0 + i), &key_ad) < 0) {
			key_ad = ICX_KEY_VALUE_DEFAULT;
		}

		key_value <<= ICX_KEY_KEY1_SFT;
		key_value |= (unsigned short)key_ad;
	}

	return key_value;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_key_save_key_value
 *  Abstruct    : Save key A/D value to SRAM
 *  Input       : key_value
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_key_save_key_value(unsigned short key_value)
{
	outl((ICX_KEY_SRAM_MAGIC | ((unsigned int)key_value & ICX_KEY_MASK_ALL)), ICX_KEY_SRAM_ADR);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_key_check_key_code
 *  Abstruct    : Check the key code
 *  Input       : none
 *  Return      : 1:equal / 0:not equal
 *----------------------------------------------------------------------------*/
int icx_key_check_key_code(unsigned short key_value, unsigned short key_code)
{
	int i, j;
	unsigned short value;
	unsigned short code = 0;
	
	for (i = 0; i < ICX_KEY_NUM; i++) {
		value = (key_value >> (i * 8)) & 0x00FF;
		for (j = 0; j < 8; j++) {
			if (value <= icx_key_value_table[j]) {
				code |= (0x1 << j) << (i * 8);
				break;
			}
		}
	}

	return ((code & key_code) == key_code);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_key_get_hold
 *  Abstruct    : Get hold ON/OFF
 *  Input       : [OUT] hold_state : 1=HOLD ON / 0=HOLD OFF
 *  Return      : period of anti-chattering
 *----------------------------------------------------------------------------*/
unsigned int icx_key_get_hold(unsigned char *hold_state)
{
	unsigned char hold_cur;
	unsigned char hold_pre = 0xFF;
	unsigned int  count = 0;

	/* read HOLD state */
	pwc_read(PWC_GPIO_IOIN1, &hold_cur);
	hold_cur &= ICX_KEY_GPIO_XHOLD;

	/* anti-chattering */
	while (hold_cur != hold_pre) {
		udelay(ICX_KEY_HOLD_CHAT_INTERVAL);
		count++;

		hold_pre = hold_cur;

		/* read HOLD state */
		pwc_read(PWC_GPIO_IOIN1, &hold_cur);
		hold_cur &= ICX_KEY_GPIO_XHOLD;
	}

	*hold_state = ((hold_cur) ? ICX_KEY_HOLD_OFF : ICX_KEY_HOLD_ON);

	return (count * ICX_KEY_HOLD_CHAT_INTERVAL);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_key_check_cancel_hold
 *  Abstruct    : Check hold ON/OFF
 *  Input       : none
 *  Return      : 
 *----------------------------------------------------------------------------*/
unsigned char icx_key_check_cancel_hold(void)
{
	unsigned char psw_state;
	unsigned char hold_state;
	unsigned int  chk_hold_period;
	int count;
	int ret = ICX_KEY_HOLD_CHK_TIMEOUT;

	/* display HOLD */
	icx_lcdc_draw_image(ICX_NVP_ID_HOLD_IMAGE);
	icx_lcdc_power_on();
	icx_lcdc_switch(ICX_LCDC_ON);
	icx_lcdc_backlight_control(ICX_LCDC_BRIGHTNESS_MAX, ICX_LCDC_FADE_INTERVAL_HOLD);

	/* check loop */
	for (count = ICX_KEY_HOLD_CHK_COUNT; count > 0; count--) {

		udelay(ICX_KEY_HOLD_CHK_INTERVAL);

		/* check ADP and VBUS */
		psw_state = icx_pm_get_psw_state();
		if (psw_state & (PWC_PSW_STATE_SDETADP | PWC_PSW_STATE_SDETUSB)) {
			ret = ICX_KEY_HOLD_CHK_CANCEL_PSW;
			break;
		}

		/* check RTC */
		if (icx_pm_check_rtc_afg() && icx_pm_check_rtc_date(NULL)) {
			ret = ICX_KEY_HOLD_CHK_CANCEL_RTC;
			break;
		}

		/* check HOLD */
		chk_hold_period = icx_key_get_hold(&hold_state);
		if (hold_state == ICX_KEY_HOLD_OFF) {
			ret = ICX_KEY_HOLD_CHK_CANCEL;
			break;
		}											/* revise the counter */
		count -= chk_hold_period / ICX_KEY_HOLD_CHK_INTERVAL;

		/* extend the checking time if KEY is pressed */
		if (icx_key_get_key_value() != ICX_KEY_NOT_PRESSED) {
			count = ICX_KEY_HOLD_CHK_COUNT + 1;
		}
	}

	/* stop LCD */
	icx_lcdc_backlight_control(ICX_LCDC_BRIGHTNESS_MIN, ICX_LCDC_FADE_INTERVAL_OFF);
	icx_lcdc_switch(ICX_LCDC_OFF);

	return ret;
}

#endif	/* CONFIG_ICX */ 
