/*
 * icx_lcdc.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "../reg.h"
#include "icx_io.h"
#include "smu.h"
#include "pwc.h"
#include "adc.h"
#include "lcdc.h"

#include "icx_nvp.h"
#include "icx_lcdc.h"


/********************/
/*@ LCDC operations */
/********************/

#ifndef CONFIG_SONY_ICX
extern icx_lcdc_ops_t lcd_nec_ops;				/* NEC reference board		*/
extern lcdc_config_t  lcdc_nec_config;
#endif
extern icx_lcdc_ops_t lcd_tmd_ltm020dl40_ops;	/* TMD LTM020DL40			*/
extern lcdc_config_t  lcdc_tmd_ltm020dl40_config;
extern icx_lcdc_ops_t lcd_tovis_r61505w_ops;	/* TOVIS R61505W			*/
extern lcdc_config_t  lcdc_tovis_r61505w_config;
extern icx_lcdc_ops_t lcd_hydis_ili9335_ops;	/* HYDIS ILI9335			*/
extern lcdc_config_t  lcdc_hydis_ili9335_config;
extern icx_lcdc_ops_t oled_sdi_s6e63d4_ops;		/* SDI S6E63D4				*/
extern lcdc_config_t  lcdc_sdi_s6e63d4_config;


static icx_lcdc_ops_t *lcdc_ops    = NULL;
static lcdc_config_t  *lcdc_config = NULL;

/* LCD-ID */
#define	ICX_LCDC_ADSEL_LCD_ID	PWC_ADSEL_ADIN8

typedef struct {
	icx_lcdc_ops_t *ops;
	lcdc_config_t  *config;
	unsigned char  lower;
	unsigned char  upper;
} icx_lcd_id_ad_t;

static const icx_lcd_id_ad_t icx_lcd_id_table[] =
{
	{	/* TOVIS / Renesus DrIC	*/
		.ops   = &lcd_tovis_r61505w_ops,
		.config= &lcdc_tovis_r61505w_config,
		.lower = 0x00,
		.upper = 0x05,
	},
	{	/* HYDIS / ILITEK  DrIC	*/
		.ops   = &lcd_hydis_ili9335_ops,
		.config= &lcdc_hydis_ili9335_config,
		.lower = 0xF7,
		.upper = 0xFF,
	},
};
#define	ICX_LCD_ID_NUM	(sizeof(icx_lcd_id_table) / sizeof(icx_lcd_id_table[0]))


/********************/
/*@ Local variables */
/********************/

static int icx_lcdc_power_on_flag = 0;
static int icx_lcdc_brightness    = 0;


/********************/
/*@ Gloval routines */
/********************/

/*-----------------------------------------------------------------------------
 *  Function    : icx_lcdc_draw_image
 *  Abstruct    : Draw image
 *  Input       : id : area ID
 *  Return      : 0(success) / others(NG)
 *----------------------------------------------------------------------------*/
int icx_lcdc_draw_image(int id)
{
	if (lcdc_config) {
		icx_nvp_get_image_data(id, 
							   (char *)lcdc_config->frameadr,
							   LCDC_FRAME_SIZE(lcdc_config->width, lcdc_config->height));
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_lcdc_backlight_control
 *  Abstruct    : Control backlight brightness
 *  Input       : deg      : percentage of brightness
 *              : interval : interval per step
 *  Return      : 0(success) / others(NG)
 *----------------------------------------------------------------------------*/
int icx_lcdc_backlight_control(int deg, int interval)
{
	int step = 1;
	int ret  = 0;

	if (lcdc_ops && lcdc_ops->backlight_control) {
		if (deg < icx_lcdc_brightness)
			step = -step;

		while (icx_lcdc_brightness != deg) {
			icx_lcdc_brightness += step;
			ret = lcdc_ops->backlight_control(lcdc_config, icx_lcdc_brightness);
			if (ret < 0) {
				icx_lcdc_brightness -= step;
				break;
			}

			udelay(interval);
		}
	}

	return ret;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_lcdc_switch
 *  Abstruct    : Start/Stop LCDC
 *  Input       : none
 *  Return      : 0(success) / others(not ready)
 *----------------------------------------------------------------------------*/
int icx_lcdc_switch(int onoff)
{
	if (icx_lcdc_power_on_flag == 0) 
		return -1;

	if (onoff == ICX_LCDC_ON) {
		lcdc_hw_bussel(LCDC_BUSSEL_LOCAL);
		lcdc_hw_start();
	}
	else {
		lcdc_hw_stop();
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_lcdc_power_on
 *  Abstruct    : Power ON LCDC
 *  Input       : none
 *  Return      : 0(success) / others(NG)
 *----------------------------------------------------------------------------*/
int icx_lcdc_power_on(void)
{
	int ret = -1;

	if (icx_lcdc_power_on_flag != 0) 
		return 0;
		
	if (lcdc_ops && lcdc_ops->power_on) {
		ret = lcdc_ops->power_on();

		if (ret == 0) {
			if (lcdc_ops->backlight_init)
				lcdc_ops->backlight_init(lcdc_config);
			
			icx_lcdc_power_on_flag = 1;
			icx_lcdc_brightness    = 0;
		}
	}

	return ret;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_lcdc_power_off
 *  Abstruct    : Power OFF LCDC
 *  Input       : none
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_lcdc_power_off(void)
{
	if (icx_lcdc_power_on_flag != 0) {
		if (lcdc_ops) {
			if (lcdc_ops->power_off) {
				lcdc_ops->power_off();
			}

			if (lcdc_ops->backlight_exit) {
				lcdc_ops->backlight_exit();
			}
		}
	}

	pwc_regulator_off(PWC_REGULATOR_ANA);
	pwc_regulator_off(PWC_REGULATOR_LOG);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_lcdc_init
 *  Abstruct    : Initialize LCD (select a device)
 *  Input       : none
 *  Return      : 0(success) / others(not found devices)
 *----------------------------------------------------------------------------*/
int icx_lcdc_init(void)
{
	/*****************/
	/* select device */
	/*****************/
#ifdef CONFIG_SONY_ICX
	unsigned int  lcd_type;
	unsigned char lcd_id;
	int i;

	/* regulator on */
	pwc_regulator_on(PWC_REGULATOR_LOG);
	udelay(10000);
	pwc_regulator_on(PWC_REGULATOR_ANA);
	udelay(10000);

	lcd_type = icx_nvp_get_sysinfo_lcd();

	switch (lcd_type) {
		case ICX_NVP_SYSINFO_LCD_BEATLE :
			lcdc_ops = &lcd_tmd_ltm020dl40_ops;	/* TMD LTM020DL40			*/
			lcdc_config = &lcdc_tmd_ltm020dl40_config;
			break;

		case ICX_NVP_SYSINFO_LCD_SWAN :
			udelay(90000);

			if (adc_read(ICX_LCDC_ADSEL_LCD_ID, &lcd_id) == 0) {
				for (i = 0; i < ICX_LCD_ID_NUM; i++) {
					if ((icx_lcd_id_table[i].lower <= lcd_id) && (lcd_id <= icx_lcd_id_table[i].upper)) {
						lcdc_ops = icx_lcd_id_table[i].ops;
						lcdc_config = icx_lcd_id_table[i].config;
						break;
					}
				}
			}
			break;

		case ICX_NVP_SYSINFO_LCD_FALCON :
			lcdc_ops = &oled_sdi_s6e63d4_ops;	/* SDI S6E63D4				*/
			lcdc_config = &lcdc_sdi_s6e63d4_config;
			break;

		default :
			return -ENODEV;
			break;
	}

#else
	lcdc_ops = &lcd_nec_ops;					/* NEC reference board		*/
	lcdc_config = &lcdc_nec_config;
#endif

	/* clear FB area */
	if (lcdc_config) {
		memset((char *)lcdc_config->frameadr, 0, LCDC_FRAME_SIZE(lcdc_config->width, lcdc_config->height));
	}

	return 0;
}


#endif	/* CONFIG_ICX */ 
