/*
 * icx_nvp.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "icx_io.h"
#include "icx_emmc.h"
#include "icx_pm.h"
#include "icx_nvp.h"
#include "icx.h"

#include "../../../../../kernel/top/include/asm-arm/arch-mp200/icx_emmc_map.h"


/********************/
/* eMMC definitions */
/********************/
#include "../emmc_mini/emmc/emmc.h"

/* boot information */
struct boot_info {
	unsigned long addr;     /* ddr address  */
	unsigned long block;    /* block number */
	unsigned long size;     /* size [byte]  */
	unsigned long erase;    /* erase flag   */
};

/* prototype */
extern void emmc_init(void);
extern int  emmc_read(struct boot_info *boot_info);


/*********************/
/*@ eMMC information */
/*********************/

/* sector information */
#define	ICX_NVP_EMMC_BASE_SECTOR	(LOADER_AREA_SECTORS + FU_LINUX_AREA_SECTORS)
#define ICX_NVP_TABLE_SECTOR		(16)

#define ICX_NVP_DATA_SECTOR_MIN		(64)
#define ICX_NVP_DATA_SECTOR_MAX		(32767)


/********************/
/*@ NVP information */
/********************/

/* area offset table */
static unsigned int icx_nvp_area_sector_table[ICX_NVP_ID_NUM] = {
	0,			/* area 00 table : u-boot parameter		*/
	8,			/* area 08 table : boot image 			*/
	9,			/* area 09 table : hold image 			*/
	10,			/* area 10 table : low battery image	*/
	1,			/* area 01 table : system parameter		*/
};


/*********************/
/*@ U-Boot parameter */
/*********************/

/* U-Boot parameter typedef */
typedef struct {
	unsigned int	sysinfo;
	unsigned char	password[32];
	unsigned int	update_flag;
	unsigned int	beep_flag;
	icx_pm_date_t	rtc_alarm;
	unsigned int	hold_mode;
} icx_nvp_uboot_parameter_t;

/* U-Boot parameter */
static icx_nvp_uboot_parameter_t icx_nvp_uboot_param;

/* U-Boot parameter area index */
enum {
	ICX_NVP_UBOOT_PARAM_SYSINFO = 0,
	ICX_NVP_UBOOT_PARAM_PASSWORD,
	ICX_NVP_UBOOT_PARAM_UPDATE,
	ICX_NVP_UBOOT_PARAM_BEEP,
	ICX_NVP_UBOOT_PARAM_RTC,
	ICX_NVP_UBOOT_PARAM_HOLD,
	ICX_NVP_UBOOT_PARAM_NUM,
};

/* U-Boot parameter info */
typedef struct {
	void *param;	/* address to variable		*/
	int	 size;		/* data size [byte]			*/
} icx_nvp_uboot_param_info_t;

static icx_nvp_uboot_param_info_t icx_nvp_uboot_param_table[ICX_NVP_UBOOT_PARAM_NUM] = {
/*   +-------------------------------------------- address to variable	*/
/*   |                                  +--------- size [byte]		 	*/
/*   |                                  |								*/
	{&icx_nvp_uboot_param.sysinfo,      4},		/* System information	*/
	{&icx_nvp_uboot_param.password[0], 32},		/* U-Boot password		*/
	{&icx_nvp_uboot_param.update_flag,  4},		/* F/W update flag		*/
	{&icx_nvp_uboot_param.beep_flag,    4},		/* Beep OK flag			*/
	{&icx_nvp_uboot_param.rtc_alarm,    8},		/* RTC alarm info		*/
	{&icx_nvp_uboot_param.hold_mode,    4},		/* Hold mode			*/
};

/* system information */
static int icx_nvp_sysinfo_mem = ICX_NVP_SYSINFO_INVALID;
static int icx_nvp_sysinfo_lcd = ICX_NVP_SYSINFO_INVALID;
static int icx_nvp_sysinfo_dac = ICX_NVP_SYSINFO_INVALID;

static int icx_nvp_ship_info = -1;


/***************/
/*@ Image data */
/***************/

#define	ICX_NVP_IMAGE_CLUSTER_NUM_MAX		32
#define	ICX_NVP_IMAGE_SECTOR_PER_CLUSTER	16

#define	c2s(c)	(c * ICX_NVP_IMAGE_SECTOR_PER_CLUSTER)


/*********************/
/*@ Flag information */
/*********************/

#define	ICX_NVP_FLAG_FW_UPDATE	0x70555766
#define	ICX_NVP_FLAG_BEEP		0x70454562
#define	ICX_NVP_FLAG_HOLD_MODE	0x64485074


/* initialized flag */
static int icx_nvp_initialized = 0;


/*******************/
/*@ Local routines */
/*******************/

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_uboot_param
 *  Abstruct    : Get small area data
 *  Input       : info   : U-Boot parameter info.
 *              : sector : sector number
 *  Return      : 0(OK) / -1(NG)
 *----------------------------------------------------------------------------*/
static int icx_nvp_get_uboot_param(icx_nvp_uboot_param_info_t *info, int sector)
{
	struct boot_info emmc_info;
	char emmc_buf[SDM_BLOCKLEN_VAL];
	int  ret;

	emmc_info.addr  = (unsigned long)&emmc_buf;
	emmc_info.block	= sector;
	emmc_info.size	= SDM_BLOCKLEN_VAL;
	emmc_info.erase = 0;

	ret = emmc_read(&emmc_info);
	if (ret != 0) {
		return -1;
	}

	/* save to info. buffer */
	memcpy(info->param, emmc_buf, info->size);

	return 0;
}

static int icx_nvp_get_system_param(void *buf, int size, int sector)
{
	struct boot_info emmc_info;
	char emmc_buf[SDM_BLOCKLEN_VAL];
	int  ret;

	emmc_info.addr  = (unsigned long)&emmc_buf;
	emmc_info.block	= sector;
	emmc_info.size	= SDM_BLOCKLEN_VAL;
	emmc_info.erase = 0;

	ret = emmc_read(&emmc_info);
	if (ret != 0) {
		return -1;
	}

	/* save to info. buffer */
	memcpy(buf, emmc_buf, size);

	return 0;
}

/********************/
/*@ Global routines */
/********************/

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_init
 *  Abstruct    : Initialize NVP driver(get U-Boot parameters)
 *  Input       : none
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_nvp_init(void)
{
	struct boot_info emmc_info;
	char emmc_buf[SDM_BLOCKLEN_VAL];
	int sector, sector_base;
	unsigned int  *sector_buf;
	int ret;
	int i;

	if (icx_nvp_initialized)
		return;

	/**************************/
	/* clear U-Boot parameter */
	/**************************/
	memset(&icx_nvp_uboot_param, 0, sizeof(icx_nvp_uboot_param));
	icx_nvp_uboot_param.sysinfo = ICX_NVP_SYSINFO_INVALID;

	icx_nvp_initialized = 1;

	/******************/
	/* get area table */
	/******************/
	sector_base = icx_emmc_get_first_sector();
	if (sector_base < 0)
		return;
	sector_base += ICX_NVP_EMMC_BASE_SECTOR;
	sector = sector_base + ICX_NVP_TABLE_SECTOR + icx_nvp_area_sector_table[ICX_NVP_ID_UBOOT_PARAM];

	emmc_info.addr  = (unsigned long)&emmc_buf;
	emmc_info.block	= sector;
	emmc_info.size	= SDM_BLOCKLEN_VAL;
	emmc_info.erase = 0;

	ret = emmc_read(&emmc_info);
	if (ret != 0) {
		return;
	}

	/************************/
	/* get U-Boot parameter */
	/************************/
	sector_buf = (unsigned int *)emmc_buf;
	for (i = 0; i < ICX_NVP_UBOOT_PARAM_NUM; i++) {
		if ((ICX_NVP_DATA_SECTOR_MIN <= sector_buf[i]) && (sector_buf[i] <= ICX_NVP_DATA_SECTOR_MAX)) {
			icx_nvp_get_uboot_param(&icx_nvp_uboot_param_table[i], (sector_base + sector_buf[i]));

			/* sysinfo analysis */
			if (i == ICX_NVP_UBOOT_PARAM_SYSINFO) {
				icx_nvp_sysinfo_mem = icx_nvp_uboot_param.sysinfo & ICX_NVP_SYSINFO_MEM_MSK;
				icx_nvp_sysinfo_lcd = icx_nvp_uboot_param.sysinfo & ICX_NVP_SYSINFO_LCD_MSK;
				icx_nvp_sysinfo_dac = icx_nvp_uboot_param.sysinfo & ICX_NVP_SYSINFO_DAC_MSK;
			}
		}
	}

	/************************/
	/* get system parameter */
	/************************/

	sector = sector_base + ICX_NVP_TABLE_SECTOR + icx_nvp_area_sector_table[ICX_NVP_ID_SYS_PARAM];

	emmc_info.addr  = (unsigned long)&emmc_buf;
	emmc_info.block	= sector;
	emmc_info.size	= SDM_BLOCKLEN_VAL;
	emmc_info.erase = 0;

	ret = emmc_read(&emmc_info);
	if (ret != 0) {
		return;
	}

	sector_buf = (unsigned int *)emmc_buf;

	if((sector_buf[ICX_NVP_SYS_SHP] >= ICX_NVP_DATA_SECTOR_MIN)
	&& (sector_buf[ICX_NVP_SYS_SHP] <= ICX_NVP_DATA_SECTOR_MAX)){
		icx_nvp_get_system_param(
			(void *)&icx_nvp_ship_info,
			sizeof(icx_nvp_ship_info),
			sector_base + sector_buf[ICX_NVP_SYS_SHP]
		);
	}

#if 0
	printf("[NVP information]\n");
	printf("sysinfo		: %08X\n", icx_nvp_uboot_param.sysinfo);
	printf("password	:");
	for (i = 0; i < 32; i++)
		printf(" %02X", icx_nvp_uboot_param.password[i]);
	printf("\n");
	printf("update		: %08X\n", icx_nvp_uboot_param.update_flag);
	printf("beep		: %08X\n", icx_nvp_uboot_param.beep_flag);
	printf("year		: %08X\n", icx_nvp_uboot_param.rtc_alarm.year);
	printf("mdhm		: %08X (%02X/%02X %02X:%02X)\n", icx_nvp_uboot_param.rtc_alarm.tm.mdhm,
														 icx_nvp_uboot_param.rtc_alarm.tm.e.month,
														 icx_nvp_uboot_param.rtc_alarm.tm.e.day,
														 icx_nvp_uboot_param.rtc_alarm.tm.e.hour,
														 icx_nvp_uboot_param.rtc_alarm.tm.e.minute);
	printf("hold		: %08X\n", icx_nvp_uboot_param.hold_mode);
#endif
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_sysinfo
 *  Abstruct    : Get system information
 *  Input       : none
 *  Return      : system information
 *----------------------------------------------------------------------------*/
unsigned int icx_nvp_get_sysinfo(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	return icx_nvp_uboot_param.sysinfo;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_sysinfo_mem
 *  Abstruct    : Get LCD device information
 *  Input       : none
 *  Return      : LCD device information
 *----------------------------------------------------------------------------*/
unsigned int icx_nvp_get_sysinfo_mem(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	if (icx_nvp_uboot_param.sysinfo == ICX_NVP_SYSINFO_INVALID)
		return ICX_NVP_SYSINFO_INVALID;
	else
		return icx_nvp_sysinfo_mem;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_sysinfo_lcd
 *  Abstruct    : Get LCD device information
 *  Input       : none
 *  Return      : LCD device information
 *----------------------------------------------------------------------------*/
unsigned int icx_nvp_get_sysinfo_lcd(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	if (icx_nvp_uboot_param.sysinfo == ICX_NVP_SYSINFO_INVALID)
		return ICX_NVP_SYSINFO_INVALID;
	else
		return icx_nvp_sysinfo_lcd;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_sysinfo_dac
 *  Abstruct    : Get DAC device information
 *  Input       : none
 *  Return      : DAC device information
 *----------------------------------------------------------------------------*/
unsigned int icx_nvp_get_sysinfo_dac(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	if (icx_nvp_uboot_param.sysinfo == ICX_NVP_SYSINFO_INVALID)
		return ICX_NVP_SYSINFO_INVALID;
	else
		return icx_nvp_sysinfo_dac;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_ship_info
 *  Abstruct    : Get Ship Information (head 4 byte)
 *  Input       : none
 *  Return      : Ship Information
 *----------------------------------------------------------------------------*/
unsigned int icx_nvp_get_ship_information(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	return icx_nvp_ship_info;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_image_data
 *  Abstruct    : Get image data
 *  Input       : id   : area ID
 *              : buf  : address to destination area
 *              : size : size of the image data
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_nvp_get_image_data(int id, char *buf, int size)
{
	struct boot_info emmc_info;
	unsigned short cluster_buf[SDM_BLOCKLEN_VAL/sizeof(short)];
	int  sector, sector_base;
	int remain;
	char *dst;
	int ret;
	int i, j;

	/******************/
	/* get area table */
	/******************/
	sector_base = icx_emmc_get_first_sector();
	if (sector_base < 0)
		return;
	sector_base += ICX_NVP_EMMC_BASE_SECTOR;
	sector = sector_base + ICX_NVP_TABLE_SECTOR + icx_nvp_area_sector_table[id];

	emmc_info.addr  = (unsigned long)&cluster_buf;
	emmc_info.block	= sector;
	emmc_info.size	= SDM_BLOCKLEN_VAL;
	emmc_info.erase = 0;

	ret = emmc_read(&emmc_info);
	if (ret != 0) {
		return;
	}

	/*******************/
	/* read image data */
	/*******************/
	dst    = buf;
	remain = size;
	for (i = 0; i < ICX_NVP_IMAGE_CLUSTER_NUM_MAX; i++) {
		sector = c2s(cluster_buf[i]);

		if ((sector < ICX_NVP_DATA_SECTOR_MIN)
		 || ((sector + ICX_NVP_IMAGE_SECTOR_PER_CLUSTER) > ICX_NVP_DATA_SECTOR_MAX)) {
			return;
		}
		sector += sector_base;

		for (j = 0; j < ICX_NVP_IMAGE_SECTOR_PER_CLUSTER; j++) {
			if (remain <= 0) {
				return;
			}

			emmc_info.addr  = (unsigned long)dst;
			emmc_info.block	= sector;

			ret = emmc_read(&emmc_info);
			if (ret != 0) {
				return;
			}

			dst = (char *)((unsigned int)dst + SDM_BLOCKLEN_VAL);
			remain -= SDM_BLOCKLEN_VAL;
			sector++;
		}
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_check_fw_update
 *  Abstruct    : Check the F/W update flag
 *  Input       : none
 *  Return      : 1:F/W update / 0:normal
 *----------------------------------------------------------------------------*/
int icx_nvp_check_fw_update(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	return (icx_nvp_uboot_param.update_flag == ICX_NVP_FLAG_FW_UPDATE);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_check_beep_flag
 *  Abstruct    : Check the beep flag
 *  Input       : none
 *  Return      : 1:Beep ON / 0:Beep OFF
 *----------------------------------------------------------------------------*/
int icx_nvp_check_beep_flag(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	return (icx_nvp_uboot_param.beep_flag != ICX_NVP_FLAG_BEEP);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_check_hold_mode
 *  Abstruct    : Check the Hold mode flag
 *  Input       : none
 *  Return      : 1:all / 0:only touch panel
 *----------------------------------------------------------------------------*/
int icx_nvp_check_hold_mode(void)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	if (icx_nvp_uboot_param.hold_mode == ICX_NVP_FLAG_HOLD_MODE)
		return ICX_NVP_HOLD_MODE_TP;
	else
		return ICX_NVP_HOLD_MODE_ALL;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_rtc_date
 *  Abstruct    : Get RTC date(year/month/day/hour/minute)
 *  Input       : date_buf : buffer address
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_nvp_get_rtc_date(icx_pm_date_t *date_buf)
{
	if (icx_nvp_initialized == 0)
		icx_nvp_init();

	date_buf->year = icx_nvp_uboot_param.rtc_alarm.year;
	date_buf->tm.mdhm = icx_nvp_uboot_param.rtc_alarm.tm.mdhm;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_nvp_get_pwd
 *  Abstruct    : Get a pwd
 *  Input       : pwd_buf : buffer address
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_nvp_get_pwd(char *pwd_buf)
{
	memcpy(pwd_buf, icx_nvp_uboot_param.password, ICX_BOOT_PWD_LEN);
}

#endif	/* CONFIG_ICX */ 
