/*
 * icx_pm.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "../reg.h"
#include "icx_io.h"
#include "pwc.h"
#include "adc.h"

#include "icx_pm.h"
#include "icx_nvp.h"
#include "icx_lcdc.h"
#include "icx_beep.h"


/****************/
/*@ Definitions */
/****************/

/* SRAM address(to save battery status) */
#define ICX_PM_SRAM_ADR_BAT			0xA0000100
#define ICX_PM_SRAM_MAGIC_BAT		0x74420000	/* "Bt" */
#define ICX_PM_SRAM_ADR_FACT1		0xA0000110
#define ICX_PM_SRAM_MAGIC_FACT1		0x31460000	/* "F1" */
#define ICX_PM_SRAM_ADR_FACT2		0xA0000114
#define ICX_PM_SRAM_MAGIC_FACT2		0x32460000	/* "F2" */
#define ICX_PM_SRAM_ADR_FACT3		0xA0000118
#define ICX_PM_SRAM_MAGIC_FACT3		0x33460000	/* "F3" */
#define ICX_PM_SRAM_ADR_WRSTFCT		0xA000011C
#define ICX_PM_SRAM_MAGIC_WRSTFCT	0x46520000	/* "RF" */

/* DDR address(battery connected flag) */
#define	ICX_PM_DDR_ADR_BATT_FLAG	0x3337EFFC
#define	ICX_PM_DDR_BATT_FLAG		0x744F6F42	/* "BoOt" */

/* threshold of LowBatt and VeryLowBatt */
#define	ICX_PM_VBAT_LOW_DEFAULT		187
#define	ICX_PM_VBAT_VLOW_DEFAULT	178

/* check LOW-BATT */
#define	ICX_PM_LOWBATT_CHK_INTERVAL	50000		/* 50ms			*/
#define	ICX_PM_LOWBATT_CHK_COUNT	200			/* 10s			*/

/* beep pattern */
#define	ICX_PM_LOWBATT_BEEP_ON1		0			/* ON  150ms	*/
#define	ICX_PM_LOWBATT_BEEP_OFF1	3			/* OFF  50ms	*/
#define	ICX_PM_LOWBATT_BEEP_ON2		4			/* ON  150ms	*/
#define	ICX_PM_LOWBATT_BEEP_OFF2	7			/* OFF 500ms	*/
#define	ICX_PM_LOWBATT_BEEP_LOOP	17

/* intc reg value */
#define INTC_IT0_IDS_MASK_ALL 0xFFFFFFFF

/* gpio reg value */
#define GIO_IDS_MASK_ALL 0xFFFFFFFF

/* pmu reg value */
#define PMU_START_SET     0x00000001
#define PMU_WDT_ENABLE    0x00000001
#define PMU_WDT_MAX_COUNT 0x0003FFFF

/* spi reg value */
#define SPI_CSW_8CLK    (7 << 12)
#define SPI_CK_DLY_ON   (1 << 2)
#define SPI_CK_POL_POS  (0 << 1)
#define SPI_CS_POL_POS  (0 << 0)
#define SPI_POL_SP0_CS0 (SPI_CSW_8CLK | SPI_CK_DLY_ON | SPI_CK_POL_POS | SPI_CS_POL_POS)

/* smu reg address for pmu */
#define SMU_CMD_POWERSW_ENA          0x0240
#define SMU_CMD_QR_ENA               0x0220
#define SMU_CMD_RESETREQ0            0x0004
#define SMU_CMD_POWERSW_ACTRL_EN     0x025c
#define SMU_CMD_L1_POWERSW           0x0244
#define SMU_CMD_L2_POWERSW           0x0500
#define SMU_CMD_LOG2SW_ACTRLEN       0x0504
#define SMU_CMD_RESETREQ2ENA         0x001c
#define SMU_CMD_RESETREQ2            0x0018
#define SMU_CMD_LOG1SW_ACTRL         0x0260
#define SMU_CMD_CLK_MODE_SEL         0x0080
#define SMU_CMD_AUTO_MODE_EN         0x007c
#define SMU_CMD_ASMU_BGCTRL          0x0208
#define SMU_CMD_RESETREQ0ENA         0x0008
#define SMU_CMD_RESETREQ1ENA         0x0010
#define SMU_CMD_POWERSW_STATUS       0x023c
#define SMU_CMD_L1_BUZ               0x0268
#define SMU_CMD_L1_BUZ2              0x026c
#define SMU_CMD_GCLKCTRL0ENA         0x01b8
#define SMU_CMD_GCLKCTRL0            0x01b4
#define SMU_CMD_GCLKCTRL1ENA         0x01c0
#define SMU_CMD_GCLKCTRL1            0x01bc
#define SMU_CMD_GCLKCTRL2ENA         0x01c8
#define SMU_CMD_GCLKCTRL2            0x01c4
#define SMU_CMD_GCLKCTRL3ENA         0x01d0
#define SMU_CMD_GCLKCTRL3            0x01cc
#define SMU_CMD_GCLKCTRL4ENA         0x0850
#define SMU_CMD_GCLKCTRL4            0x084c
#define SMU_CMD_RESETREQ3            0x083c
#define SMU_CMD_RESETREQ3ENA         0x0840
#define SMU_CMD_RESETREQ1            0x000c
#define SMU_CMD_PLL1CTRL1            0x0088
#define SMU_CMD_PLL3CTRL1            0x0098
#define SMU_CMD_DIVSP0SCLK           0x0118
#define SMU_CMD_IO_L0_LM_BUZ         0x0814
#define SMU_CMD_MEMC_HAND_SHAKE_FAKE 0x08a0
#define SMU_CMD_AHBCLKCTRL0          0x01a0
#define SMU_CMD_AHBCLKCTRL1          0x01a4
#define SMU_CMD_APBCLKCTRL0          0x01a8
#define SMU_CMD_APBCLKCTRL1          0x01ac

/* chg reg address for pmu */
#define CHG_CMD_CHG_L1_HOLD 0x0004

/* memc reg address for pmu */
#define MEMC_CMD_MEMC_DDR_STATE8   0x202c
#define MEMC_CMD_MEMC_CACHE_MODE   0x0000
#define MEMC_CMD_MEMC_DEGFUN       0x0008
#define MEMC_CMD_MEMC_REQSCH       0x1000
#define MEMC_CMD_MEMC_DDR_CONFIGF  0x2000
#define MEMC_CMD_MEMC_DDR_CONFIGA1 0x2004
#define MEMC_CMD_MEMC_DDR_CONFIGA2 0x2008
#define MEMC_CMD_MEMC_DDR_CONFIGR1 0x2014
#define MEMC_CMD_MEMC_DDR_CONFIGR2 0x2018
#define MEMC_CMD_MEMC_DDR_CONFIGR3 0x201c
#define MEMC_CMD_MEMC_DDR_CONFIGT1 0x2020
#define MEMC_CMD_MEMC_DDR_CONFIGT2 0x2024
#define MEMC_CMD_MEMC_DDR_CONFIGC1 0x200c
#define MEMC_CMD_MEMC_DDR_CONFIGC2 0x2010

/* spi reg address for pmu */
#define SPI_CMD_SPx_TX_DATA 0x0010
#define SPI_CMD_SPx_CONTROL 0x0008
#define SPI_CMD_SPx_FFCLR   0x0028
#define SPI_CMD_SPx_ENSET   0x0020
#define SPI_CMD_SPx_MODE    0x0000
#define SPI_CMD_SPx_POL     0x0004

/* pmu pc */
#define PMU_PC_MAIN                0x1000
#define PMU_PC_MEMC_DEEP_POWERDOWN 0x1110
#define PMU_PC_POWEROFF            0x1400
#define PMU_PC_SPI_ENABLE          0x1450
#define PMU_PC_SPI_DISABLE         0x14a0
#define PMU_PC_POWER_ON_SEQUENCE   0x1500
#define PMU_PC_L2POWEROFF          0x1800
#define PMU_PC_MEMC1               0x1900
#define PMU_PC_L1POWEROFF          0x1b00
#define PMU_PC_USE_MEMC            0x1f00

/* smu macro value */
#define ATIM_MACRO    (0x00 << 16)
#define AINT_MACRO    (0x01 << 16)
#define LCD_MACRO     (0x02 << 16)
#define GPIO_MACRO    (0x03 << 16)
#define SI1_MACRO     (0x04 << 16)
#define MEMC_MACRO    (0x05 << 16)
#define PWM_MACRO     (0x06 << 16)
#define PMU_MACRO     (0x09 << 16)
#define ASMU_MACRO    (0x0A << 16)
#define SPI0_MACRO    (0x0B << 16)
#define SPI1_MACRO    (0x0C << 16)
#define CHG_MACRO     (0x0D << 16)
#define SI0_MACRO     (0x0E << 16)
#define AXL0_MACRO    (0x0F << 16)
#define ASMU_S2_MACRO (0x11 << 16)
#define AINT_S2_MACRO (0x12 << 16)
#define MEMC_S2_MACRO (0x13 << 16)

/* pmu command parameters */
#define CMD_REG_WRITE        (0x00 << 25)
#define CMD_SP0_WRITE        (0x01 << 25)
#define CMD_REG_READ         (0x02 << 25)
#define CMD_RWM              (0x03 << 25)
#define CMD_MOVE             (0x04 << 25)
#define CMD_AND              (0x05 << 25)
#define CMD_EXOR             (0x06 << 25)
#define CMD_CMP1             (0x07 << 25)
#define CMD_CMP2             (0x08 << 25)
#define CMD_REG_WRITE2       (0x09 << 25)
#define CMD_BRANCH           (0x10 << 25)
#define CMD_JUMP             (0x11 << 25)
#define CMD_AJUMP            (0x12 << 25)
#define CMD_SUBROUTINE_START (0x13 << 25)
#define CMD_RFS              (0x14 << 25)
#define CMD_TIMERWAIT        (0x20 << 25)
#define CMD_INTWAIT          (0x21 << 25)
#define CMD_CCPU_STAT_WAIT   (0x22 << 25)
#define CMD_SMU_READY_WAIT   (0x23 << 25)
#define CMD_TRIG_WAIT        (0x24 << 25)
#define CMD_QR_WAIT          (0x25 << 25)
#define CMD_CYCLE_WAIT       (0x26 << 25)
#define CMD_LCD_MODE_WAIT    (0x27 << 25)
#define CMD_INT_MASK         (0x30 << 25)
#define CMD_ARMINT_MASK      (0x31 << 25)
#define CMD_WDT_CLEAR        (0x32 << 25)
#define CMD_WDT_STOP         (0x33 << 25)
#define CMD_WDT_RESTART      (0x34 << 25)
#define CMD_PMU_END          (0x35 << 25)
#define CMD_NOP              (0x36 << 25)

/* pmu other parameters */
#define PMU_REGA                  (0 << 24)
#define PMU_REGB                  (1 << 24)
#define PMU_BEQ                   (0 << 24)
#define PMU_BNE                   (1 << 24)
#define PMU_WAIT_INT_END          (0 << 24)
#define PMU_WAIT_INT_CONTINUE     (1 << 24)
#define PMU_LOW_LEVEL             (0 << 24)
#define PMU_HIGH_LEVEL            (1 << 24)
#define PMU_TRIG_TIMER            (1 << 19)
#define PMU_TRIG_CCPU_ACCESS_LOW  (1 << 20)
#define PMU_TRIG_CCPU_ACCESS_HIGH (1 << 21)
#define PMU_TRIG_CCPU_LOWPWR_LOW  (1 << 22)
#define PMU_TRIG_CCPU_LOWPWR_HIGH (1 << 23)
#define PMU_TRIG_INT              (1 << 24)
#define PMU_QR_WAIT_BACKUP        (0 << 24)
#define PMU_QR_WAIT_RESTORE       (1 << 24)
#define PMU_INT_UNMASK            (0 << 24)
#define PMU_INT_MASK              (1 << 24)

/* rtc bit asiign */
#define	PWC_RTC_RTC1_WALE		0x80
#define	PWC_RTC_RTC1_ALE		0xC0
#define	PWC_RTC_RTC2_AFG		0x03
#define	PWC_RTC_RTC2_ADJ		0x10
#define	PWC_INTC_WRSTMSK_RTCAL	0x01

/* pmu macro */
#define PMU_VIRADDR_TO_PC(addr) (addr - PMU_BASE)
#define PMU_PC_TO_VIRADDR(pc)   (pc + PMU_BASE)
#define PMU_CMD_REG_WRITE(macro, addr, data, idx) \
{ \
    outl((CMD_REG_WRITE | macro | addr), idx++); \
    outl(data, idx++); \
}
#define PMU_CMD_SP0_WRITE(addr, data, idx) \
{ \
    outl((CMD_SP0_WRITE | addr), idx++); \
    outl(data, idx++); \
}
#define PMU_CMD_REG_READ(macro, addr, aorb, idx) \
{ \
    outl((CMD_REG_READ | aorb | macro | addr), idx++); \
}
#define PMU_CMD_RWM(macro, addr, data, data_en, idx) \
{ \
    outl((CMD_RWM | macro | addr), idx++); \
    outl(data, idx++); \
    outl(data_en, idx++); \
}
#define PMU_CMD_MOVE(data, aorb, idx) \
{ \
    outl((CMD_MOVE | aorb), idx++); \
    outl(data, idx++); \
}
#define PMU_CMD_AND(data, aorb, idx) \
{ \
    outl((CMD_AND | aorb), idx++); \
    outl(data, idx++); \
}
#define PMU_CMD_EXOR(data, aorb, idx) \
{ \
    outl((CMD_EXOR | aorb), idx++); \
    outl(data, idx++); \
}
#define PMU_CMD_CMP1(idx) \
{ \
    outl(CMD_CMP1, idx++); \
}
#define PMU_CMD_CMP2(data, aorb, idx) \
{ \
    outl((CMD_CMP2 | aorb), idx++); \
    outl(data, idx++); \
}
#define PMU_CMD_REG_WRITE2(macro, addr, aorb, idx) \
{ \
    outl((CMD_REG_WRITE2 | aorb | macro | addr),idx++); \
}
#define PMU_CMD_BRANCH(jump, op, idx) \
{ \
    outl((CMD_BRANCH | op | jump), idx++); \
}
#define PMU_CMD_JUMP(jump, idx) \
{ \
    outl((CMD_JUMP | jump), idx++); \
}
#define PMU_CMD_AJUMP(jump, idx) \
{ \
    outl((CMD_AJUMP | jump), idx++); \
}
#define PMU_CMD_SUBROUTINE_START(jump, idx) \
{ \
    outl((CMD_SUBROUTINE_START | jump), idx++); \
}
#define PMU_CMD_RFS(idx) \
{ \
    outl(CMD_RFS, idx++); \
}
#define PMU_CMD_TIMERWAIT(count, int, idx) \
{ \
    outl((CMD_TIMERWAIT | int | count), idx++); \
}
#define PMU_CMD_INTWAIT(idx) \
{ \
    outl(CMD_INTWAIT, idx++); \
}
#define PMU_CMD_CCPU_STAT_WAIT(level, idx) \
{ \
    outl((CMD_CCPU_STAT_WAIT | level), idx++); \
}
#define PMU_CMD_SMU_READY_WAIT(level, idx) \
{ \
    outl((CMD_SMU_READY_WAIT | level), idx++); \
}
#define PMU_CMD_TRIG_WAIT(count, trig, idx) \
{ \
    outl((CMD_TRIG_WAIT | trig | count), idx++); \
}
#define PMU_CMD_QR_WAIT(borr, idx) \
{ \
    outl((CMD_QR_WAIT | borr), idx++); \
}
#define PMU_CMD_CYCLE_WAIT(count, int, idx) \
{ \
    outl((CMD_CYCLE_WAIT | int | count), idx++); \
}
#define PMU_CMD_LCD_MODE_WAIT(idx) \
{ \
    outl(CMD_LCD_MODE_WAIT, idx++); \
}
#define PMU_CMD_INT_MASK(mask, idx) \
{ \
    outl((CMD_INT_MASK | mask), idx++); \
}
#define PMU_CMD_ARMINT_MASK(mask, idx) \
{ \
    outl((CMD_ARMINT_MASK | mask), idx++); \
}
#define PMU_CMD_WDT_CLEAR(idx) \
{ \
    outl(CMD_WDT_CLEAR, idx++); \
}
#define PMU_CMD_WDT_STOP(idx) \
{ \
    outl(CMD_WDT_STOP, idx++); \
}
#define PMU_CMD_WDT_RESTART(idx) \
{ \
    outl(CMD_WDT_RESTART, idx++); \
}
#define PMU_CMD_PMU_END(idx) \
{ \
    outl(CMD_PMU_END, idx++); \
}
#define PMU_CMD_NOP(idx) \
{ \
    outl(CMD_NOP, idx++); \
}

/***************/
/*@ Prototypes */
/***************/
static void setup_pmu(void);
static void set_pmu_poweroff_sequence(void);
/* static void mp200_cpu_do_idle(void); */

/********************/
/*@ Global routines */
/********************/

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_init
 *  Abstruct    : Initialize PM driver
 *  Input       : none
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_pm_init(void)
{
	outl(ICX_PM_SRAM_MAGIC_BAT, ICX_PM_SRAM_ADR_BAT);

	/* outl(ICX_PM_SRAM_MAGIC_FACT1, ICX_PM_SRAM_ADR_FACT1); */
	/* outl(ICX_PM_SRAM_MAGIC_FACT2, ICX_PM_SRAM_ADR_FACT2); */
	/* outl(ICX_PM_SRAM_MAGIC_FACT3, ICX_PM_SRAM_ADR_FACT3); */
	/* outl(ICX_PM_SRAM_MAGIC_WRSTFCT, ICX_PM_SRAM_ADR_WRSTFCT); */
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_get_boot_factor
 *  Abstruct    : Get boot factor
 *  Input       : none
 *  Return      : boot factor
 *				:	[-		:bit31-24]	-
 *				:	[FACTOR3:bit23-16]	bit17: FRTCINT
 *				:	[FACTOR2:bit15- 8]	-
 *				:	[FACTOR1:bit 7- 0]	bit 7: ACC_DET
 *				:						bit 1: KEY_XHOLD
 *				:						bit 0: KEY_WAKE
 *----------------------------------------------------------------------------*/
unsigned int icx_pm_get_boot_factor(void)
{
	unsigned char factor1, factor2, factor3, wrstfct;

	/* get FACTOR */
	if (pwc_read(PWC_INTC_FACTOR1, &factor1) < 0)
		factor1 = 0;
	if (pwc_read(PWC_INTC_FACTOR2, &factor2) < 0)
		factor2 = 0;
	if (pwc_read(PWC_INTC_FACTOR3, &factor3) < 0)
		factor3 = 0;
	if (pwc_read(PWC_INTC_WRSTFCT, &wrstfct) < 0)
		wrstfct = 0;

	/* clear FACTOR */
	pwc_write(PWC_INTC_CLRFACT1, PWC_INTC_FACTOR_ALL, PWC_INTC_FACTOR_ALL);
	pwc_write(PWC_INTC_CLRFACT2, PWC_INTC_FACTOR_ALL, PWC_INTC_FACTOR_ALL);
	pwc_write(PWC_INTC_CLRFACT3, PWC_INTC_FACTOR_ALL, PWC_INTC_FACTOR_ALL);
	pwc_write(PWC_INTC_WRSTCL, PWC_INTC_WRSTFCT_ALL,PWC_INTC_WRSTFCT_ALL);

	/* save to SRAM */
	outl(ICX_PM_SRAM_MAGIC_FACT1  |factor1, ICX_PM_SRAM_ADR_FACT1); 
	outl(ICX_PM_SRAM_MAGIC_FACT2  |factor2, ICX_PM_SRAM_ADR_FACT2); 
	outl(ICX_PM_SRAM_MAGIC_FACT3  |factor3, ICX_PM_SRAM_ADR_FACT3); 
	outl(ICX_PM_SRAM_MAGIC_WRSTFCT|wrstfct, ICX_PM_SRAM_ADR_WRSTFCT); 

	/* clear RTCINT if AFG is not set */
	if (icx_pm_check_rtc_afg() == 0) 
		factor3 &= ~PWC_INTC_FACTOR3_FRTCINT;

	return (((unsigned int)factor3 << ICX_PM_FACTOR3_SFT) | (unsigned int)factor1);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_get_psw_state
 *  Abstruct    : Get PSW state
 *  Input       : none
 *  Return      : PSW state
 *----------------------------------------------------------------------------*/
unsigned char icx_pm_get_psw_state(void)
{
	unsigned char state;

	/* get PSW state */
	if (pwc_read(PWC_PSW_STATE, &state) < 0)
		state = 0;

	return state;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_get_vbat
 *  Abstruct    : Get VCCVBAT
 *  Input       : none
 *  Return      : VBAT
 *----------------------------------------------------------------------------*/
unsigned char icx_pm_get_vbat(void)
{
	unsigned char vbat;

	/* get VBAT */
	if (adc_read(PWC_ADSEL_VCCVBAT, &vbat) < 0)
		vbat = 0;

	return vbat;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_get_vbat_threshold
 *  Abstruct    : Get VBAT threshold of LowBatt and VeryLowBatt
 *  Input       : low_batt  (OUT) : threshold of LowBatt
 *				: vlow_batt (OUT) : threshold of VeryLowBatt
 *  Return      : 0:success / -EINVAL
 *----------------------------------------------------------------------------*/
int icx_pm_get_vbat_threshold(unsigned char *low_batt, unsigned char *vlow_batt)
{
	if ((low_batt == NULL) || (vlow_batt == NULL))
		return -EINVAL;

	*low_batt  = ICX_PM_VBAT_LOW_DEFAULT;
	*vlow_batt = ICX_PM_VBAT_VLOW_DEFAULT;

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_check_battery_flag
 *  Abstruct    : Check the battery flag
 *  Input       : none
 *  Return      : 1:connected / 0:not connected
 *----------------------------------------------------------------------------*/
int icx_pm_check_battery_flag(void)
{
	if (inl(ICX_PM_DDR_ADR_BATT_FLAG) == ICX_PM_DDR_BATT_FLAG)
		return 1;
	else
		return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_set_battery_flag
 *  Abstruct    : Set the battery flag
 *  Input       : 1:set the flag / 0:clear the flag
 *  Return		: none
 *----------------------------------------------------------------------------*/
void icx_pm_set_battery_flag(int flag)
{
	if (flag == ICX_PM_BATT_FLAG_SET)
		outl(ICX_PM_DDR_BATT_FLAG, ICX_PM_DDR_ADR_BATT_FLAG);
	else
		outl(0x00000000, ICX_PM_DDR_ADR_BATT_FLAG);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_check_cancel_vbat_low
 *  Abstruct    : Check the vbat whether LOW or not
 *  Input       : none
 *  Return      : 1:canceled LOWBATT / 0:LOWBATT
 *----------------------------------------------------------------------------*/
int icx_pm_check_cancel_vbat_low(void)
{
	int count;
	int beep_pattern;
	unsigned char psw_state;
	int ret = ICX_PM_LOWBATT_CHK_TIMEOUT;

	/* display LowBatt */
    icx_lcdc_draw_image(ICX_NVP_ID_LOWBATT_IMAGE);
	icx_lcdc_power_on();
	icx_lcdc_switch(ICX_LCDC_ON);
	icx_lcdc_backlight_control(ICX_LCDC_BRIGHTNESS_MAX, ICX_LCDC_FADE_INTERVAL_LOWBATT);

	/* enable BEEP */
	icx_beep_enable();

	for (count = 0; count < ICX_PM_LOWBATT_CHK_COUNT; count++) {

		beep_pattern = count % ICX_PM_LOWBATT_BEEP_LOOP;

		/* check to connect ADP or VBUS */
		if ((beep_pattern == 0) || (beep_pattern > ICX_PM_LOWBATT_BEEP_OFF2)) {
			psw_state = icx_pm_get_psw_state();
			if (psw_state & (PWC_PSW_STATE_SDETADP | PWC_PSW_STATE_SDETUSB)) {
				ret = ICX_PM_LOWBATT_CHK_CANCEL;
				break;
			}
		}

		/* sound a beep */
		switch (beep_pattern) {
			case ICX_PM_LOWBATT_BEEP_ON1 :
			case ICX_PM_LOWBATT_BEEP_ON2 :
				/* sound a beep */
				icx_beep_on(ICX_BEEP_TIME_CONTINUOUS);
				break;
			case ICX_PM_LOWBATT_BEEP_OFF1 :
			case ICX_PM_LOWBATT_BEEP_OFF2 :
				/* stop the beep */
				icx_beep_off();
				break;
			default :
				break;
		}

		udelay(ICX_PM_LOWBATT_CHK_INTERVAL);
	}

	/* write black data */
	lcdc_hw_bussel(LCDC_BUSSEL_BLACK);

	udelay(40000);

	/* stop LCD */
	icx_lcdc_backlight_control(ICX_LCDC_BRIGHTNESS_MIN, ICX_LCDC_FADE_INTERVAL_OFF);
	icx_lcdc_switch(ICX_LCDC_OFF);

	return ret;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_check_rtc_afg
 *  Abstruct    : Check AFG flag
 *  Input       : none
 *  Return      : 1:AFG flag ON / 0:AFG flag OFF
 *----------------------------------------------------------------------------*/
int icx_pm_check_rtc_afg(void)
{
	unsigned char rtc2_rd;
	int ret = 0;

	if (pwc_read(PWC_RTC_RTC2_RD, &rtc2_rd) == 0) {
		if (rtc2_rd & PWC_RTC_RTC2_AFG) {
			ret = 1;
		}
	}

	return ret;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_get_rtc_date
 *  Abstruct    : Get date(year/month/day/hour/minute)
 *  Input       : date_buf : buffer address
 *  Return      : 0:success / -EINVAL / -EIO
 *----------------------------------------------------------------------------*/
int icx_pm_get_rtc_date(icx_pm_date_t *date_buf)
{
	unsigned char data;
	int ret = 0;

	if (date_buf == NULL)
		return -EINVAL;

	ret |= pwc_read(PWC_RTC_YEAR_RD, &data);
	date_buf->year = data;
	ret |= pwc_read(PWC_RTC_MONTH_RD, &data);
	date_buf->tm.e.month = data;
	ret |= pwc_read(PWC_RTC_DAY_RD, &data);
	date_buf->tm.e.day = data;
	ret |= pwc_read(PWC_RTC_HOUR_RD, &data);
	date_buf->tm.e.hour = data;
	ret |= pwc_read(PWC_RTC_MIN_RD, &data);
	date_buf->tm.e.minute = data;

#if 0
	printf("[RTC alarm]\n");
	printf("year        : %08X\n", date_buf->year);
	printf("mdhm        : %08X (%02X/%02X %02X:%02X)\n", date_buf->tm.mdhm,
														 date_buf->tm.e.month,
														 date_buf->tm.e.day,
														 date_buf->tm.e.hour,
														 date_buf->tm.e.minute);
#endif

	return ret;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_check_rtc_date
 *  Abstruct    : Check RTC date(year/month/day/hour/minute)
 *  Input       : date_buf : buffer address (NULL:get current date)
 *  Return      : 0:not equal / 1:equal
 *----------------------------------------------------------------------------*/
int icx_pm_check_rtc_date(icx_pm_date_t *date_buf)
{
	icx_pm_date_t date_now;
	icx_pm_date_t date_nvp;

	/* read alarm time from NVP */
	icx_nvp_get_rtc_date(&date_nvp);

	if (date_buf == NULL) {
		/* read current date */
		if (icx_pm_get_rtc_date(&date_now) != 0)
			return 0;

		/* compare current date with NVP */
		if ((date_now.year == date_nvp.year) && (date_now.tm.mdhm == date_nvp.tm.mdhm))
			return 1;
	}
	else {
		/* compare specified date(date_buf) with NVP */
		if ((date_buf->year == date_nvp.year) && (date_buf->tm.mdhm == date_nvp.tm.mdhm))
			return 1;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_set_vbus_limit
 *  Abstruct    : set vbus limit
 *  Input       : limit value 500 100 0
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_pm_set_vbus_limit(int limit)
{
	if(limit==0){
		/* VBUS_SUSPEND : high */
		pwc_write(PWC_GPIO_IOOUT2,0x40,0x48);
	}
	else if(limit==100){
#ifdef INVERT_VBUS_I_HIGH
		/* VBUS_SUSPEND : low  */
		/* VBUS_I_LOW   : high */
		pwc_write(PWC_GPIO_IOOUT2,0x08,0x48);
#else
		/* VBUS_SUSPEND : low */
		/* VBUS_I_HIGH  : low */
		pwc_write(PWC_GPIO_IOOUT2,0x00,0x48);
#endif
	}
	else{ /* 500 */
#ifdef INVERT_VBUS_I_HIGH
		/* VBUS_SUSPEND : low */
		/* VBUS_I_LOW   : low */
		pwc_write(PWC_GPIO_IOOUT2,0x00,0x48);
#else
		/* VBUS_SUSPEND : low  */
		/* VBUS_I_HIGH  : high */
		pwc_write(PWC_GPIO_IOOUT2,0x08,0x48);
#endif
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_set_charge_switch
 *  Abstruct    : set charge enable/disabe
 *  Input       : 1/0
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_pm_set_charge_swicth(int flag)
{
	if(flag){
#ifdef INVERT_CHG_XEN
		/* CHG_EN : high */
		pwc_write(PWC_GPIO_IOOUT2,0x10,0x10);
#else
		/* CHG_XEN : low */
		pwc_write(PWC_GPIO_IOOUT2,0x00,0x10);
#endif
	}
	else{
#ifdef INVERT_CHG_XEN
		/* CHG_EN : low */
		pwc_write(PWC_GPIO_IOOUT2,0x00,0x10);
#else
		/* CHG_XEN : high */
		pwc_write(PWC_GPIO_IOOUT2,0x10,0x10);
#endif
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_set_charge_voltage
 *  Abstruct    : set charge volatge
 *  Input       : charge voltage 90/100
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_pm_set_charge_voltage(int voltage)
{
	if(voltage==90){
		/* CHG_V_LOW : high  */
		pwc_write(PWC_GPIO_IOOUT2,0x02,0x02);
	}
	else{
		/* CHG_V_LOW : low  */
		pwc_write(PWC_GPIO_IOOUT2,0x02,0x02);
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_save_status_sram
 *  Abstruct    : Save the battery status to SRAM
 *  Input       : battery status
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_pm_save_status_sram(unsigned int status)
{
	outl(ICX_PM_SRAM_MAGIC_BAT|status, ICX_PM_SRAM_ADR_BAT);
}

/*-----------------------------------------------------------------------------
 *  Function    : icx_pm_power_off
 *  Abstruct    : Power off the system
 *  Input       : none
 *  Return      : none
 *----------------------------------------------------------------------------*/
void icx_pm_power_off(void)
{
	unsigned char reg;
	int i;

	/*********************/
	/* mask all interupt */
	/*********************/

	/* disbale fiq */
	/* u-boot does not enable fiq. */

	/* disable irq */
	/* u-boot does not enable irq. */

	/* disable int */
	outl(INTC_IT0_IDS_MASK_ALL,INTC_IT0_IDS0);
	outl(INTC_IT0_IDS_MASK_ALL,INTC_IT0_IDS1);
	outl(INTC_IT0_IDS_MASK_ALL,INTC_IT0_IDS2);

	/* disable gpio int */
	outl(GIO_IDS_MASK_ALL,GIO_IDS_L);
	outl(GIO_IDS_MASK_ALL,GIO_IDS_H);
	outl(GIO_IDS_MASK_ALL,GIO_IDS_HH);
	outl(GIO_IDS_MASK_ALL,GIO_IDS_HHH);

	/* disable pwc gpio */
	pwc_write(PWC_INTC_BMASK1,0x00,0xFF);
	pwc_write(PWC_INTC_BMASK2,0x00,0xFF);
	pwc_write(PWC_INTC_BMASK3,0x00,0xFF);

	/****************/
	/* pull control */
	/****************/

	/* HP_XMUTE2    pull-down 0x40 */
	/* VBUS_SUSPEND pull-none 0x00 */
	/* CHG_I_HIGH   pull-none 0x00 */
	/* CHG_XEN      pull-none 0x00 */
	pwc_write(PWC_GPIO_GPPUPD4,0x40,0xFF);

	/* VBUS_I_HIGH  pull-none 0x00 */
	/* PWC_GPIO_10  pull-down 0x10 */
	/* CHG_V_LOW    pull-none 0x00 */
	/* PWC_GPIO_08  pull-down 0x01 */
	pwc_write(PWC_GPIO_GPPUPD3,0x11,0xFF);

	/* ACC_DET      pull-down 0x40 */
	/* CHG_STAT1    pull-none 0x00 */
	/* CHG_STAT2    pull-none 0x00 */
	/* PWC_GPIO_04  pull-down 0x01 */
	pwc_write(PWC_GPIO_GPPUPD2,0x41,0xFF);

	/* PWC_GPIO_03  pull-down 0x40 */
	/* KEY_NC_ON    pull-down 0x10 */
	/* KEY_XHOLD    pull-none 0x00 */
	/* KEY_WAKE     pull-none 0x00 */
	pwc_write(PWC_GPIO_GPPUPD1,0x50,0xFF);

	/****************/
	/* output value */
	/****************/

#ifdef INVERT_VBUS_I_HIGH
#ifdef INVERT_CHG_XEN
	/* HP_XMUTE2    low  0x00 */
	/* VBUS_SUSPEND low  0x00 */
	/* CHG_I_HIGH   high 0x20 */
	/* CHG_EN       low  0x00 */
	/* VBUS_I_LOW   low  0x00 */
	/* PWC_GPIO_10  low  0x00 */
	/* CHG_V_LOW    low  0x00 */
	/* PWC_GPIO_08  low  0x00 */
	pwc_write(PWC_GPIO_IOOUT2,0x20,0xFF);
#else
	/* HP_XMUTE2    low  0x00 */
	/* VBUS_SUSPEND low  0x00 */
	/* CHG_I_HIGH   high 0x20 */
	/* CHG_XEN      high 0x10 */
	/* VBUS_I_LOW   low  0x00 */
	/* PWC_GPIO_10  low  0x00 */
	/* CHG_V_LOW    low  0x00 */
	/* PWC_GPIO_08  low  0x00 */
	pwc_write(PWC_GPIO_IOOUT2,0x30,0xFF);
#endif
#else
#ifdef INVERT_CHG_XEN
	/* HP_XMUTE2    low  0x00 */
	/* VBUS_SUSPEND low  0x00 */
	/* CHG_I_HIGH   high 0x20 */
	/* CHG_EN       low  0x00 */
	/* VBUS_I_HIGH  high 0x08 */
	/* PWC_GPIO_10  low  0x00 */
	/* CHG_V_LOW    low  0x00 */
	/* PWC_GPIO_08  low  0x00 */
	pwc_write(PWC_GPIO_IOOUT2,0x28,0xFF);
#else
	/* HP_XMUTE2    low  0x00 */
	/* VBUS_SUSPEND low  0x00 */
	/* CHG_I_HIGH   high 0x20 */
	/* CHG_XEN      high 0x10 */
	/* VBUS_I_HIGH  high 0x08 */
	/* PWC_GPIO_10  low  0x00 */
	/* CHG_V_LOW    low  0x00 */
	/* PWC_GPIO_08  low  0x00 */
	pwc_write(PWC_GPIO_IOOUT2,0x38,0xFF);
#endif
#endif

	/* ACC_DET      low  0x00 */
	/* CHG_STAT1    low  0x00 */
	/* CHG_STAT2    low  0x00 */
	/* PWC_GPIO_04  low  0x00 */
	/* PWC_GPIO_03  low  0x00 */
	/* KEY_NC_ON    low  0x00 */
	/* KEY_XHOLD    low  0x00 */
	/* KEY_WAKE     low  0x00 */
	pwc_write(PWC_GPIO_IOOUT1,0x00,0xFF);

	/******************/
	/* port direction */
	/******************/

	/* HP_XMUTE2    in  0x00 */
	/* VBUS_SUSPEND out 0x40 */
	/* CHG_I_HIGH   out 0x20 */
	/* CHG_XEN      out 0x10 */
	/* VBUS_I_HIGH  out 0x08 */
	/* PWC_GPIO_10  in  0x00 */
	/* CHG_V_LOW    out 0x02 */
	/* PWC_GPIO_08  in  0x00 */
	pwc_write(PWC_GPIO_IOSEL2,0x7A,0xFF);

	/* ACC_DET      in  0x00 */
	/* CHG_STAT1    in  0x00 */
	/* CHG_STAT2    in  0x00 */
	/* PWC_GPIO_04  in  0x00 */
	/* PWC_GPIO_03  in  0x00 */
	/* KEY_NC_ON    in  0x00 */
	/* KEY_XHOLD    in  0x00 */
	/* KEY_WAKE     in  0x00 */
	pwc_write(PWC_GPIO_IOSEL1,0x00,0xFF);

	/*************************/
	/* interrupt detect mode */
	/*************************/

	/* USB_DET_INT  level 0x00 */
	/* ADP_DET_INT  level 0x00 */
	/* BAT_DET_INT  level 0x00 */
	/* USB WAKE_INT level 0x00 */
	/* pwc_write(PWC_INTC_DETMOD3H,0x00,0xFF); */

	/* RTC_INT      rise  0x04 */
	pwc_write(PWC_INTC_DETMOD3L,0x04,0x0C);

	/* ACC_DET      rise  0x40 */
	pwc_write(PWC_INTC_DETMOD1H,0x40,0xC0);

	/* KEY_XHOLD    both  0x04 */
	/* KEY_WAKE     rise  0x01 */
	pwc_write(PWC_INTC_DETMOD1L,0x05,0x0F);

	/******************************/
	/* device side interrupt mask */
	/******************************/

	/* USB_DET_INT  mask    0x00 */
	/* ADP_DET_INT  mask    0x00 */
	/* BAT_DET_INT  mask    0x00 */
	/* USB WAKE_INT mask    0x00 */
	/* POWEROFF_INT mask    0x00 */
	/* CHARGE_INT   mask    0x00 */
	/* RTC_INT      through 0x02 */
	/* AD_INT       mask    0x00 */
	pwc_write(PWC_INTC_FMASK3,0x02,0xFF);

	/* ACC_DET      through 0x80 */
	/* CHG_STAT1    mask    0x00 */
	/* CHG_STAT2    mask    0x00 */
	/* PWC_GPIO_04  mask    0x00 */
	/* PWC_GPIO_03  mask    0x00 */
	/* KEY_NC_ON    mask    0x00 */
	/* KEY_XHOLD    through 0x02 */
	/* KEY_WAKE     through 0x01 */
	pwc_write(PWC_INTC_FMASK1,0x83,0xFF);

	/****************/
	/* clear factor */
	/****************/

	/* rtc int reset */
	pwc_read(PWC_RTC_RTC1_RD,&reg);
	if(reg&PWC_RTC_RTC1_ALE){
		pwc_write(PWC_RTC_RTC1,(reg&~PWC_RTC_RTC1_ALE),0xFF);
		pwc_write(PWC_RTC_RTC1,reg,0xFF);
	}

	/* wait 1ms */
	for (i=12*1000; i>0; i--)
		;

	/* clear all factor */
	pwc_write(PWC_INTC_CLRFACT1,0xFF,0xFF);
	pwc_write(PWC_INTC_CLRFACT2,0xFF,0xFF);
	pwc_write(PWC_INTC_CLRFACT3,0xFF,0xFF);

	/**************/
	/* power off  */
	/**************/

	/* setup pmu for poweroff */
	setup_pmu();

	/* setup rtc poweron register */
	pwc_read(PWC_RTC_RTC1_RD,&reg);
	if(reg&PWC_RTC_RTC1_ALE)
		pwc_write(PWC_INTC_WRSTMSK,0x01,0xFF);
	else
		pwc_write(PWC_INTC_WRSTMSK,0x00,0xFF);

	/* ACC_DET      through 0x80 */
	/* CHG_STAT1    mask    0x00 */
	/* CHG_STAT2    mask    0x00 */
	/* PWC_GPIO_04  mask    0x00 */
	/* PWC_GPIO_03  mask    0x00 */
	/* KEY_NC_ON    mask    0x00 */
	/* KEY_XHOLD    through 0x02 */
	/* KEY_WAKE     through 0x01 */
	pwc_write(PWC_INTC_REGPWONMSK,0x8F,0xFF);

	/* need at power off ? */
	/* power switch wait param = 0x53 */
	/* armnormal = normal             */
	/* quick recovery = enable        */
	outl(0x00530101,ASMU_QR_ENA);

	/* goto idle(wfi) and wait power off */
	cpu_idle();
}

static void setup_pmu()
{
	/* unsigned int regval; */
	unsigned int val;

	/* pmu command sequence set */
	set_pmu_poweroff_sequence();

	/* set start pc */
	outl(PMU_PC_MAIN,PMU_PC);

	/* set start pc at power on */
	outl(PMU_PC_POWER_ON_SEQUENCE,PMU_POWER_ON_PC);

	/* set pmu watch dog */
	outl(PMU_WDT_ENABLE,PMU_WDT_COUNT_EN);
	outl(PMU_WDT_MAX_COUNT,PMU_WDT_COUNT_LMT);

	/* pmu start */
	outl(PMU_START_SET, PMU_START);
	while(1){
		val=inl(PMU_START);
		if(val&PMU_START_SET)
			break;
	}
}

static void set_pmu_poweroff_sequence()
{
	unsigned int* pmu_cmd_adr;
	unsigned tmp_pc;

	/* PMU_CMD_REG_WRITE(GPIO_MACRO, 0x000C, 0x80008000, pmu_cmd_adr); */
	/* PMU_CMD_REG_WRITE(GPIO_MACRO, 0x0048, 0x00010001, pmu_cmd_adr); */
	/* PMU_CMD_REG_WRITE(GPIO_MACRO, 0x008C, 0x40004000, pmu_cmd_adr); */

	/******************************************************************/
	/********** PowerOffSequence (Sleep/DeepSleep/PowerOff) ***********/
	/******************************************************************/

	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_MAIN);

	/* M-1 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_POWERSW_ENA, 0x000fff05, pmu_cmd_adr);
	/* M-2 */
	PMU_CMD_RWM(ASMU_MACRO, SMU_CMD_QR_ENA, 0x00000002, 0x00000002, pmu_cmd_adr);
	PMU_CMD_QR_WAIT(PMU_QR_WAIT_BACKUP, pmu_cmd_adr);
	/* M-3 */
	PMU_CMD_INT_MASK(PMU_INT_MASK, pmu_cmd_adr);

	/* M-5 */
	PMU_CMD_REG_WRITE(CHG_MACRO, CHG_CMD_CHG_L1_HOLD, 0x00000001, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(CHG_MACRO, CHG_CMD_CHG_L1_HOLD, 0x00000003, pmu_cmd_adr);

	/* M-6 */ 
	PMU_CMD_SUBROUTINE_START(PMU_PC_MEMC_DEEP_POWERDOWN, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_IO_L0_LM_BUZ, 0x00000000, pmu_cmd_adr);
	PMU_CMD_SUBROUTINE_START(PMU_PC_MEMC1, pmu_cmd_adr);
	PMU_CMD_SUBROUTINE_START(PMU_PC_L1POWEROFF, pmu_cmd_adr);
	PMU_CMD_SUBROUTINE_START(PMU_PC_L2POWEROFF, pmu_cmd_adr);

	/* M-7 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_MEMC_HAND_SHAKE_FAKE, 0x00000001, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_RESETREQ2ENA, 0x0000000a, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_RESETREQ2, 0x00000000, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_RESETREQ2ENA, 0x00000000, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* M-8 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_CLK_MODE_SEL, 0x00000005, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_SMU_READY_WAIT(PMU_HIGH_LEVEL, pmu_cmd_adr);

	/* M-9 */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(ASMU_MACRO, SMU_CMD_CLK_MODE_SEL, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000f00, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000500, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* M-10 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_CLK_MODE_SEL, 0x00000006, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_SMU_READY_WAIT(PMU_HIGH_LEVEL, pmu_cmd_adr);

	/* M-11 */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(ASMU_MACRO, SMU_CMD_CLK_MODE_SEL, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000f00, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000600, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* M-12 */
	PMU_CMD_AJUMP(PMU_PC_POWEROFF , pmu_cmd_adr);

	/***************************/
	/******** POWEROFF *********/
	/***************************/

	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_POWEROFF);

	/* PWOFF-1 */
	PMU_CMD_SUBROUTINE_START(PMU_PC_SPI_ENABLE, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(SPI0_MACRO, SPI_CMD_SPx_TX_DATA, 0x0000021c, pmu_cmd_adr);
	PMU_CMD_SP0_WRITE(SPI_CMD_SPx_CONTROL, 0x00000009, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(SPI0_MACRO, SPI_CMD_SPx_FFCLR, 0x000000ff, pmu_cmd_adr);
	PMU_CMD_SUBROUTINE_START(PMU_PC_SPI_DISABLE, pmu_cmd_adr);
	/* PWOFF-2 */
	PMU_CMD_INTWAIT(pmu_cmd_adr)


	/***************************/
	/******** SPI Enable *******/
	/***************************/

	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_SPI_ENABLE);

	/* SPIEN-1 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_GCLKCTRL3ENA, 0x00600000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_GCLKCTRL3, 0x00600000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_GCLKCTRL3ENA, 0x00000000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(SPI0_MACRO, SPI_CMD_SPx_FFCLR, 0x000000ff, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(SPI0_MACRO, SPI_CMD_SPx_ENSET, 0x000000ff, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(SPI0_MACRO, SPI_CMD_SPx_MODE, 0x00000f00, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(SPI0_MACRO, SPI_CMD_SPx_POL, SPI_POL_SP0_CS0, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_DIVSP0SCLK, 0x00000190, pmu_cmd_adr);
	PMU_CMD_RFS(pmu_cmd_adr);

	/***************************/
	/******** SPI Disable ******/
	/***************************/

	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_SPI_DISABLE);

	/* SPIDIS-1 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_GCLKCTRL3ENA, 0x00600000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_GCLKCTRL3, 0x00000000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_GCLKCTRL3ENA, 0x00000000, pmu_cmd_adr);
	PMU_CMD_RFS(pmu_cmd_adr);

	/******************************/
	/******** L2OFF_Sequence ******/
	/******************************/

	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_L2POWEROFF);

	/* L2OFF-1 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_LOG2SW_ACTRLEN, 0x00000100, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_L2_POWERSW, 0x00000000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_L2_POWERSW, 0x000000FF, pmu_cmd_adr);
	PMU_CMD_RFS(pmu_cmd_adr);

	/*******************************/
	/******** MEMC1_Sequence *******/
	/*******************************/

	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_MEMC1);

	/* MEMC1_SEQ1 */
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_CACHE_MODE, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0000), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DEGFUN, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0004), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_REQSCH, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x000c), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGF, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0010), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGA1, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0014), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGA2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0018), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGR1, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x001c), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGR2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0020), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGR3, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0024), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGT1, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0028), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGT2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x002c), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC1, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_REG_WRITE2(PMU_MACRO, (PMU_PC_USE_MEMC + 0x0030), PMU_REGB, pmu_cmd_adr);
	PMU_CMD_RFS(pmu_cmd_adr);

	/***********************************/
	/******** PMU_PC_L1POWEROFF ********/
	/***********************************/
	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_L1POWEROFF);

	/* L1OFF-1 */
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_L1_BUZ, 0x00000000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_L1_BUZ2, 0x00000000, pmu_cmd_adr);
	PMU_CMD_REG_WRITE(ASMU_MACRO, SMU_CMD_L1_POWERSW, 0x00000F0F, pmu_cmd_adr);
	PMU_CMD_RFS(pmu_cmd_adr);

	/**************************************/
	/********* MEMC_DEEP_POWERDOWN ********/
	/**************************************/

	pmu_cmd_adr = (unsigned int*)PMU_PC_TO_VIRADDR(PMU_PC_MEMC_DEEP_POWERDOWN);

	/* MEMC DPD-0 auto power down:disable */
	PMU_CMD_RWM(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGR3, 0x00000000, 0x00000101, pmu_cmd_adr); 
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* MEMC_DPD-1 Auto-SREF,Auto-CBR:disable */ 
	PMU_CMD_RWM(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGR2, 0x00000000, 0x00000303, pmu_cmd_adr); 
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* CMD CLEAR */

	/* ready */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* write */
	PMU_CMD_REG_WRITE(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, 0x000000c0, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* read */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000000, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* MEMC_DPD-2 all bank precharge */
	PMU_CMD_RWM(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC1, 0x00000400, 0x00000400, pmu_cmd_adr); 

	/* ready */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* write */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_WRITE(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, 0x00000099, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* read */
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000009, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* CMD CLEAR */

	/* ready */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* write */
	PMU_CMD_REG_WRITE(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, 0x000000c0, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* read */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000000, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* MEMC_DPD-3 CKE */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);

	/* ready */
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* write */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_WRITE(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, 0x0000009d, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* read */
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x0000000d, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* CMD CLEAR */

	/* ready */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* write */
	PMU_CMD_REG_WRITE(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, 0x000000c0, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* read */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000000, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* MEMC_DPD-5 Deep Power Down */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);

	/* ready */
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* write */
	PMU_CMD_REG_WRITE(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, 0x0000009c, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* read */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x0000000c, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* check state */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_STATE8, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000007, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* CMD CLEAR */

	/* ready */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000300, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	/* write */
	PMU_CMD_REG_WRITE(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, 0x000000c0, pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);
	PMU_CMD_NOP(pmu_cmd_adr);

	/* read */
	tmp_pc = PMU_VIRADDR_TO_PC((unsigned int)pmu_cmd_adr);
	PMU_CMD_REG_READ(MEMC_MACRO, MEMC_CMD_MEMC_DDR_CONFIGC2, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_AND(0x0000000f, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_CMP2(0x00000000, PMU_REGB, pmu_cmd_adr);
	PMU_CMD_BRANCH(tmp_pc, PMU_BNE, pmu_cmd_adr);

	PMU_CMD_RFS(pmu_cmd_adr);

	return;
}

#endif	/* CONFIG_ICX */ 
