/*
 * lcd_nec.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifndef CONFIG_SONY_ICX

#include <asm/errno.h>

#include "icx_io.h"
#include "smu.h"
#include "spi.h"
#include "gpio.h"
#include "lcdc.h"

#include "icx_lcdc.h"

/*******************/
/*@ LCDC operation */
/*******************/

static int lcd_nec_power_on(void);
static int lcd_nec_backlight_control(lcdc_config_t *config, int deg);

icx_lcdc_ops_t lcd_nec_ops = {
	.power_on		   = lcd_nec_power_on,
	.power_off		   = NULL,

	.backlight_init    = NULL,
	.backlight_exit    = NULL,
	.backlight_control = lcd_nec_backlight_control,
};


/******************/
/*@ Local setting */
/******************/

/* hsync */
#define LCD_NEC_HAREA		LCDC_WIDTH_WVGA
#define	LCD_NEC_HPULSE		1
#define	LCD_NEC_HFRONTP		6
#define	LCD_NEC_HBACKP		4
#define	LCD_NEC_HBLANK		(LCD_NEC_HFRONTP + LCD_NEC_HPULSE + LCD_NEC_HBACKP)
#define	LCD_NEC_HTOTAL		(LCD_NEC_HAREA + LCD_NEC_HBLANK)
#define	LCD_NEC_HEDGE1		(LCD_NEC_HFRONTP)
#define	LCD_NEC_HEDGE2		(LCD_NEC_HFRONTP + LCD_NEC_HPULSE)

/* vsync */
#define LCD_NEC_VAREA		LCDC_HEIGHT_WVGA
#define	LCD_NEC_VPULSE		1
#define	LCD_NEC_VFRONTP		3
#define	LCD_NEC_VBACKP		6
#define	LCD_NEC_VBLANK		(LCD_NEC_VFRONTP + LCD_NEC_VPULSE + LCD_NEC_VBACKP)
#define	LCD_NEC_VTOTAL		(LCD_NEC_VAREA + LCD_NEC_VBLANK)
#define	LCD_NEC_VEDGE1		(LCD_NEC_VFRONTP)
#define	LCD_NEC_VEDGE2		(LCD_NEC_VFRONTP + LCD_NEC_VPULSE)

/* LCDC config */
lcdc_config_t lcdc_nec_config = {
	.width	  = LCDC_WIDTH_WVGA,
	.height   = LCDC_HEIGHT_WVGA,

	.clkdiv   = SMU_PLLSEL_PLL2 | SMU_DIV20,

	.frameadr = LCDC_FRAMEADR_DEFAULT,
    .areaadr  = LCDC_AREAADR_DEFAULT,
	.pol	  = (LCDC_CLKPOL_POS | LCDC_HPOL_POS | LCDC_VPOL_POS | LCDC_ENPOL_HI),
	.htotal   = LCD_NEC_HTOTAL, 
	.harea	  = LCD_NEC_HAREA,
	.hedge1   = LCD_NEC_HEDGE1, 
	.hedge2   = LCD_NEC_HEDGE2,
	.vtotal   = LCD_NEC_VTOTAL,
	.varea    = LCD_NEC_VAREA,
	.vedge1   = LCD_NEC_VEDGE1,
	.vedge2   = LCD_NEC_VEDGE2,
};

/* SPI config */
static spi_config_t spi_config_w = {
	.dev = SPI_DEV_SP0,
	.clk = SPI_CLK_3MHZ,
	.cs  = SPI_CS_SEL_CS1,
	.pol = (SPI_CK_DLY_ON | SPI_CK_POL_POS | SPI_CS_POL_NEG),
	.nbw = SPI_NB_32BIT,
	.nbr = 0,
};

/* SPI command */
static unsigned int spi_cmd_power_on[][2] = {
	{0x03, 0x01}, {0x00, 0x00}, {0x01, 0x01}, {0x04, 0x00}, {0x05, 0x14},
	{0x06, 0x24}, {0x10, 0xd7}, {0x11, 0x00}, {0x12, 0x00}, {0x13, 0x55},
	{0x14, 0x01}, {0x15, 0x70}, {0x16, 0x1e}, {0x17, 0x25}, {0x18, 0x25},
	{0x19, 0x02}, {0x1a, 0x02}, {0x1b, 0xa0}, {0x20, 0x2f}, {0x21, 0x0f},
	{0x22, 0x0f}, {0x23, 0x0f}, {0x24, 0x0f}, {0x25, 0x0f}, {0x26, 0x0f},
	{0x27, 0x00}, {0x28, 0x02}, {0x29, 0x02}, {0x2a, 0x02}, {0x2b, 0x0f},
	{0x2c, 0x0f}, {0x2d, 0x0f}, {0x2e, 0x0f}, {0x2f, 0x0f}, {0x30, 0x0f},
	{0x31, 0x0f}, {0x32, 0x00}, {0x33, 0x02}, {0x34, 0x02}, {0x35, 0x02},
	{0x50, 0x0c}, {0x53, 0x42}, {0x54, 0x42}, {0x55, 0x41}, {0x56, 0x14},
	{0x59, 0x88}, {0x5a, 0x01}, {0x5b, 0x00}, {0x5c, 0x02}, {0x5d, 0x0c},
	{0x5e, 0x1c}, {0x5f, 0x27}, {0x62, 0x49}, {0x63, 0x27}, {0x66, 0x76},
	{0x67, 0x27}, {0x70, 0x01}, {0x71, 0x0e}, {0x72, 0x02}, {0x73, 0x0c},
	{0x76, 0x0c}, {0x79, 0x30}, {0x82, 0x00}, {0x83, 0x00}, {0x84, 0xfc},
	{0x86, 0x00}, {0x88, 0x00}, {0x8a, 0x00}, {0x8b, 0x00}, {0x8c, 0x00},
	{0x8d, 0xfc}, {0x8f, 0x00}, {0x91, 0x00}, {0x93, 0x00}, {0x94, 0x00},
	{0x95, 0x00}, {0x96, 0xfc}, {0x98, 0x00}, {0x9a, 0x00}, {0x9c, 0x00},
	{0x9d, 0x00}, {0x02, 0x00}
};

#define	SPI_CMD_POWER_ON_NUM	(sizeof(spi_cmd_power_on)/sizeof(spi_cmd_power_on[0]))

/* GPIO */
#define	GPIO_LCD_RST		GPIO_P3
#define	GPIO_LCD_PD			GPIO_P2


/*******************/
/*@ Local routines */
/*******************/

/*-----------------------------------------------------------------------------
 *  Function    : lcd_nec_power_on
 *  Abstruct    : Power on a LCD on NEC reference board
 *  Input       : config : pointer to LCDC configure[OUT]
 *  Return      : 0(success) / others(NG)
 *----------------------------------------------------------------------------*/
static int lcd_nec_power_on(void)
{
	unsigned int spi_data;
	int i;
	int ret;

	/* H/W unreset */
	gpio_set_value(GPIO_LCD_RST, 0x1);
	gpio_set_direction(GPIO_LCD_RST, GPIO_OUT);

	/* H/W power on */
	spi_probe(&spi_config_w);
	for (i = 0; i < SPI_CMD_POWER_ON_NUM; i++) {
		spi_data = (spi_cmd_power_on[i][0] << 16) | (1 << 8) | spi_cmd_power_on[i][1];
		ret = spi_write(&spi_config_w, &spi_data, 1);
		if (ret < 0) {
			printf("NEC LCD :spi_write error(%d)\n", ret);
			return ret;
		}
	}

	/* LCDC power on */
	lcdc_hw_init();

	/* LCDC initialize */
	lcdc_hw_set(&lcdc_nec_config);

	/* backlight setting */
	gpio_set_value(GPIO_LCD_PD, 0x0);
	gpio_set_direction(GPIO_LCD_PD, GPIO_OUT);

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : lcd_nec_backlight_control
 *  Abstruct    : Control backlight brightness for NEC LCD
 *  Input       : LCDC configure
 *				: deg : percentage of brightness(0-100)
 *  Return      : 0(success) / -EPERM
 *----------------------------------------------------------------------------*/
static int lcd_nec_backlight_control(lcdc_config_t *config, int deg)
{
	if ((deg < 0) || (deg > 100))
		return -EPERM;

	if (deg == 0)
		gpio_set_value(GPIO_LCD_PD, 0x0);
	else
		gpio_set_value(GPIO_LCD_PD, 0x1);

	return 0;
}

#endif	/* !CONFIG_SONY_ICX */ 
