/*
 * pwc.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "../reg.h"
#include "icx_io.h"
#include "pwc.h"
#include "spi.h"
#include "gpio.h"

/****************/
/*@ Address map */ 
/****************/

/* for mirror of PWC register */
#define	PWC_REG_MAGIC_NO		0x63495770		/* pWIc */
#define PWC_REG_MAGIC_ADR		0xA0000000
#define	PWC_REG_SRAM_BASE		0xA0000004
#define PWC_REG_NUM				0x100

/***********/
/*@ define */
/***********/

#define CHG_PINSEL_G32_KM_PWRON1_MSK	0x03000000
#define CHG_PINSEL_G32_KM_PWRON1		0x00000000
#define CHG_PINSEL_G32_KM_PWRON2_MSK	0x0C000000
#define CHG_PINSEL_G32_KM_PWRON2		0x00000000
#define CHG_PULL_G40_KM_PWRON1_MSK		0x00070000
#define CHG_PULL_G40_KM_PWRON1			0x00050000
#define CHG_PULL_G40_KM_PWRON2_MSK		0x00700000
#define CHG_PULL_G40_KM_PWRON2			0x00500000

/********************/
/*@ Type definition */
/********************/

/* reset value */
typedef struct {
	u_char addr;
	u_char data;
} pwc_reg_data_t;


/********************/
/*@ Local variables */
/********************/

/* mirror of PWC registers */
static u_char *pwc_reg_mirror = (u_char *)PWC_REG_SRAM_BASE;
static u_char  pwc_reg_rd[PWC_REG_NUM];

/* table of initial vaule */
static pwc_reg_data_t pwc_reset_data[] = {
	{PWC_PWR_REGA,	 	0x03},
	{PWC_PWR_PSCNT2,	0x02},
	{PWC_PWR_VOUTSET1,	0xAC},
	{PWC_PWR_VOUTSET2,	0xA8},
	{PWC_GPIO_GPPUPD3, 	0x55},
	{PWC_GPIO_GPPUPD4, 	0x55},
	{PWC_CHG_CHGCNTR1, 	0x03},
	{PWC_CHG_TIMCNTR, 	0x02},
	{PWC_USB_USBCONT, 	0x07},
};

#define	PWC_REG_RESET_NUM	(sizeof(pwc_reset_data)/sizeof(pwc_reset_data[0]))

/* table of readable registers */
static u_char pwc_reg_rd_tbl[] = {
	PWC_GPIO_IOIN1,
	PWC_GPIO_IOIN2,
	PWC_INTC_FACTOR1,
	PWC_INTC_FACTOR2,
	PWC_INTC_FACTOR3,
	PWC_ADC_ADDATA1,
	PWC_ADC_ADDATA2,
	PWC_ADC_ADDATA3,
	PWC_ADC_ADDATA4,
	PWC_ADC_ADDATA5,
	PWC_ADC_ADDATA6,
	PWC_ADC_ADDATA7,
	PWC_ADC_ADDATA8,
	PWC_CHG_CHGIRR,
	PWC_CHG_CHGSTATE,
	PWC_PSW_STATE,
	PWC_INTC_WRSTFCT,
	PWC_LSIVER,
	PWC_RTC_SEC_RD,
	PWC_RTC_MIN_RD,
	PWC_RTC_HOUR_RD,
	PWC_RTC_WEEK_RD,
	PWC_RTC_DAY_RD,
	PWC_RTC_MONTH_RD,
	PWC_RTC_YEAR_RD,
	PWC_RTC_TRIM_RD,
	PWC_RTC_ALM_WM_RD,
	PWC_RTC_ALM_WH_RD,
	PWC_RTC_ALM_WW_RD,
	PWC_RTC_ALM_DM_RD,
	PWC_RTC_ALM_DH_RD,
	PWC_RTC_RTC1_RD,
	PWC_RTC_RTC2_RD,
};

#define PWC_REG_RD_NUM      (sizeof(pwc_reg_rd_tbl)/sizeof(pwc_reg_rd_tbl[0]))

static int pwc_regulator_count[PWC_MAX_REGULATOR]={0};

/********/
/*@ SPI */
/********/

/* SPI config */
static spi_config_t spi_config_w = {
	.dev = SPI_DEV_SP0,
	.clk = SPI_CLK_24MHZ,
	.cs  = SPI_CS_SEL_CS0,
	.pol = (SPI_CK_DLY_ON | SPI_CK_POL_POS | SPI_CS_POL_POS),
	.nbw = SPI_NB_16BIT,
	.nbr = 0,
};

static spi_config_t spi_config_r = {
	.dev = SPI_DEV_SP0,
	.clk = SPI_CLK_12MHZ,
	.cs  = SPI_CS_SEL_CS0,
	.pol = (SPI_CK_DLY_ON | SPI_CK_POL_POS | SPI_CS_POL_POS),
	.nbw = SPI_NB_8BIT,
	.nbr = SPI_NB_8BIT,
};


/********************/
/*@ Global routines */
/********************/

/*-----------------------------------------------------------------------------
 *  Function    : pwc_init
 *  Abstruct    : Initialize PowerIC
 *  Input       :
 *  Return      : 
 *----------------------------------------------------------------------------*/
void pwc_init(void)
{
	u_int regval;
	int i;

	/* set magic number */
	outl(PWC_REG_MAGIC_NO, PWC_REG_MAGIC_ADR);

	/* clear mirror region */
	memset(pwc_reg_mirror, 0, PWC_REG_NUM);

	/* set reset values */
	for (i = 0; i < PWC_REG_RESET_NUM; i++) {
		pwc_reg_mirror[pwc_reset_data[i].addr] = pwc_reset_data[i].data;
	}

	/* clear readable flag table */
	memset(pwc_reg_rd, 0, PWC_REG_NUM);

	/* set readable flag */
	for (i = 0; i < PWC_REG_RD_NUM; i++) {
		pwc_reg_rd[pwc_reg_rd_tbl[i]] = 1;
	}

#ifndef CONFIG_SONY_ICX	/* only NEC reference board */
	pwc_reg_mirror[PWC_PWR_PSCNT2]  = 0x0F; 
	pwc_reg_mirror[PWC_GPIO_IOSEL1] = 0x80;
	pwc_reg_mirror[PWC_GPIO_IOOUT1] = 0x00;
#endif /* !CONFIG_SONY_ICX */

#ifdef CONFIG_SONY_ICX
	/* Clear PWONMSK */
	pwc_write(PWC_INTC_REGPWONMSK,0x00,0xFF);

	/* RegUSB Off */
	pwc_write(PWC_USB_USBCONT,0x05,0xFF);

	/*************************/
	/* GPIO default settings */
	/*************************/

	/* all bit low */
	pwc_write(PWC_GPIO_IOOUT1, 0x00,0xFF);

	/* vbus limit: 500 / charge: off */
#ifdef INVERT_VBUS_I_HIGH
#ifdef INVERT_CHG_XEN
	/* CHG_I_HIGH  : 0x20 */
	pwc_write(PWC_GPIO_IOOUT2, 0x20,0xFF);
#else
	/* CHG_I_HIGH  : 0x20 */
	/* CHG_XEN     : 0x10 */
	pwc_write(PWC_GPIO_IOOUT2, 0x30,0xFF);
#endif
#else
#ifdef INVERT_CHG_XEN
	/* CHG_I_HIGH  : 0x20 */
	/* VBUS_I_HIGH : 0x08 */
	pwc_write(PWC_GPIO_IOOUT2, 0x28,0xFF);
#else
	/* CHG_I_HIGH  : 0x20 */
	/* CHG_XEN     : 0x10 */
	/* VBUS_I_HIGH : 0x08 */
	pwc_write(PWC_GPIO_IOOUT2, 0x38,0xFF);
#endif
#endif

	/* all bit input */
	pwc_write(PWC_GPIO_IOSEL1,0x00,0xFF);

	/* VBUS_SUSPEND : output */
	/* CHG_I_HIGH   : output */
	/* CHG_XEN      : output */
	/* VBUS_I_HIGH  : output */
	/* CHG_V_LOW    : output */
	pwc_write(PWC_GPIO_IOSEL2,0x7A,0xFF);

	/* set gpio pull (default down) */

	/* all bit pull-down */
	pwc_write(PWC_GPIO_GPPUPD1,0x55,0xFF);

	/* all bit pull-down */
	pwc_write(PWC_GPIO_GPPUPD2,0x55,0xFF);

	/* CHG_V_LOW   : pull-none */
	/* VBUS_I_HIGH : pull-none */
	pwc_write(PWC_GPIO_GPPUPD3,0x11,0xFF);

	/* VBUS_SUSPEND : pull-none */
	/* CHG_I_HIGH   : pull-none */
	/* CHG_XEN      : pull-none */
	pwc_write(PWC_GPIO_GPPUPD4,0x40,0xFF);

	/* for regulator on/off */
	regval = inl(CHG_PINSEL_G32) & ~(CHG_PINSEL_G32_KM_PWRON1_MSK | CHG_PINSEL_G32_KM_PWRON2_MSK);
	outl((regval | (CHG_PINSEL_G32_KM_PWRON1 | CHG_PINSEL_G32_KM_PWRON2)), CHG_PINSEL_G32);
	gpio_set_value(BEEP_GIO_KM_PWRON1, 0x0);
	gpio_set_direction(BEEP_GIO_KM_PWRON1, GPIO_OUT);
	gpio_set_value(BEEP_GIO_KM_PWRON2, 0x0);
	gpio_set_direction(BEEP_GIO_KM_PWRON2, GPIO_OUT);
	regval = inl(CHG_PULL_G40) & ~(CHG_PULL_G40_KM_PWRON1_MSK | CHG_PULL_G40_KM_PWRON2_MSK);
	outl((regval | (CHG_PULL_G40_KM_PWRON1 | CHG_PULL_G40_KM_PWRON2)), CHG_PULL_G40);

#endif /* CONFIG_SONY_ICX */
}

/*-----------------------------------------------------------------------------
 *  Function    : pwc_regulator_on
 *  Abstruct    : regulator on
 *  Input       : regualtor : regulator define
 *  Return      : 0         : success
 *                minus     : error
 *----------------------------------------------------------------------------*/
int pwc_regulator_on(unsigned int regulator)
{
	unsigned short addr;
	unsigned short data;
	unsigned short mask;
	int regu;

	addr=(unsigned short)(regulator&0xFF);
	data=0xFFFF;
	mask=(unsigned short)((regulator>>8)&0xFF);
	regu=(int)((regulator>>16)&0xFF);

	if(addr!=PWC_PWR_PSCNT1 && addr!=PWC_PWR_PSCNT2 && addr!=PWC_USB_USBCONT)
		return(-1);

	if(regu<0 || regu>=PWC_MAX_REGULATOR)
		return(-1);

	if(addr==PWC_USB_USBCONT){
		data=data|0x0005;
		mask=mask|0x0005;
	}

	if(regulator==PWC_REGULATOR_HP){
		gpio_set_value(BEEP_GIO_KM_PWRON2,1);
		udelay(50000);
	}

	/* regulator on */
	pwc_write(addr,data,mask);

	if(regulator==PWC_REGULATOR_ANA){
		udelay(1000);
		gpio_set_value(BEEP_GIO_KM_PWRON1,1);
	}

	pwc_regulator_count[regu]++;

	return(0);
}

/*-----------------------------------------------------------------------------
 *  Function    : pwc_regulator_off
 *  Abstruct    : regulator off
 *  Input       : regualtor : regulator define
 *  Return      : 0         : success
 *                minus     : error
 *----------------------------------------------------------------------------*/
int pwc_regulator_off(unsigned int regulator)
{
	unsigned short addr;
	unsigned short data;
	unsigned short mask;
	int regu;

	addr=(unsigned short)(regulator&0xFF);
	data=0x0000;
	mask=(unsigned short)((regulator>>8)&0xFF);
	regu=(int)((regulator>>16)&0xFF);

	if(addr!=PWC_PWR_PSCNT1 && addr!=PWC_PWR_PSCNT2 && addr!=PWC_USB_USBCONT)
		return(-1);

	if(regu<0 || regu>=PWC_MAX_REGULATOR)
		return(-1);

	if(addr==PWC_USB_USBCONT){
		data=data|0x0005;
		mask=mask|0x0005;
	}

	if(pwc_regulator_count[regu]<=1){

		if(regulator==PWC_REGULATOR_ANA){
			gpio_set_value(BEEP_GIO_KM_PWRON1,0);
			udelay(1000);
		}

		/* regulator off */
		pwc_write(addr,data,mask);

		if(regulator==PWC_REGULATOR_HP){
			udelay(50000);
			gpio_set_value(BEEP_GIO_KM_PWRON2,0);
		}

	}

	if(pwc_regulator_count[regu]>0)
		pwc_regulator_count[regu]--;

	return(0);
}

/*-----------------------------------------------------------------------------
 *  Function    : pwc_read
 *  Abstruct    : Read a data from register of PowerIC
 *  Input       : addr      : address of register
 *                data      : data
 *  Return      : 0         : success(read from PowerIC)
 *                1			: success(read from mirror)
 *                -EINVAL   : invalid parameter
 *----------------------------------------------------------------------------*/
int pwc_read(u_int addr, u_char *data)
{
	u_int spi_data;
	int   ret;

	if (addr >= PWC_REG_NUM)
		return -EINVAL;

	if (pwc_reg_rd[addr]) {
		ret = 0;

		/* SPI read */
    	spi_probe(&spi_config_r);
		ret = spi_read(&spi_config_r, &addr, &spi_data, SPI_READ_1CYCLE);
		*data = (u_char)spi_data;

		return ret;
	}
	else {
		*data = pwc_reg_mirror[addr];
		return 1;
	}
}

/*-----------------------------------------------------------------------------
 *  Function    : pwc_write
 *  Abstruct    : Write a data to register of PowerIC
 *  Input       : addr      : address of register
 *                data      : data
 *                mask      : mask
 *  Return      : 0         : success
 *                -EINVAL   : invalid parameter
 *----------------------------------------------------------------------------*/
int pwc_write(u_int addr, u_char data, u_char mask)
{
	u_char tmp_data;
	u_int  buf;
	int    ret;

	/* read current value */
	ret = pwc_read(addr, &tmp_data);
	if (ret <= 0)
		return ret;

	/* update value and write to PowerIC */
    spi_probe(&spi_config_w);
	tmp_data = (tmp_data & ~mask) | (data & mask);
	buf = (u_int)(tmp_data & 0xFF) | (addr << 8);
	ret = spi_write(&spi_config_w, &buf, 1);
	if (ret < 0)
		return ret;

	/* update mirror */
	pwc_reg_mirror[addr] = tmp_data;

	return 0;
}

#endif	/* CONFIG_ICX */ 
