/*
 * pwm.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include "icx_io.h"
#include "../reg.h"
#include "gpio.h"
#include "smu.h"
#include "pwm.h"


/******************/
/*@ Local setting */
/******************/

/* pinsel */
#define CHG_PINSEL_G80_MSK_PWM1		0xC0000000
#define CHG_PINSEL_G80_PWM1			0x40000000

#define CHG_PULL_G88_MSK_PWM1		0x70000000
#define CHG_PULL_G88_PWM1			0x50000000

/* clock divisor */
#define SMU_DIVPWMPWCLK_INIT_MSK	0x0000FFFF
#define SMU_DIVPWMPWCLK_INIT		(SMU_DIV20 << 16)

/* PWM setting */
#define PWM_MODE					0x00000001	/* enable CMP0		*/
#define PWM_DELAY					5			/* Delay 5 cycles	*/
#define PWM_LEDGE					0
#define PWM_TEDGE					(PWM_LEDGE + 1)

#define PWM_START					0x00000001
#define PWM_STOP					0x00000000


/******************/
/*@ Local routine */
/******************/

/*-----------------------------------------------------------------------------
 *  Function    : pwm_start
 *	Abstract	: Start PWM
 *  Input       : none
 *  Return      : none
 *	Note		: use only PWM1(28.672MHz same as kernel)
 *----------------------------------------------------------------------------*/
static void pwm_start(void)
{
	if (!(inl(PWM_CH1_CTRL) & PWM_START))
		outl(PWM_START, PWM_CH1_CTRL);
}

/*-----------------------------------------------------------------------------
 *  Function    : pwm_stop
 *	Abstract	: Stop PWM
 *  Input       : none
 *  Return      : none
 *	Note		: use only PWM1(28.672MHz same as kernel)
 *----------------------------------------------------------------------------*/
static void pwm_stop(void)
{
	outl(PWM_STOP, PWM_CH1_CTRL);
}


/*******************/
/*@ Global routine */
/*******************/

/*-----------------------------------------------------------------------------
 *  Function    : pwm_init
 *	Abstract	: Initialize PWM
 *  Input       : config : LCDC configure
 *  Return      : none
 *	Note		: use only PWM1(11.4688MHz)
 *----------------------------------------------------------------------------*/
void pwm_init(lcdc_config_t *config)
{
	u_int regval;

	/* BLON pinsel */
	regval = inl(CHG_PINSEL_G64) & ~CHG_PINSEL_G64_MSK_BLON;
	outl((regval | CHG_PINSEL_G64_BLON), CHG_PINSEL_G64);

	/* BLON pull control */
	regval = inl(CHG_PULL_G72) & ~CHG_PULL_G72_MSK_BLON;
	outl((regval | CHG_PULL_G72_BLON), CHG_PULL_G72);

	/* set BLON output */
	gpio_set_direction(LCDC_GIO_LCD_BLON, GPIO_OUT);

	/* pinsel */
	regval = inl(CHG_PINSEL_G80) & ~CHG_PINSEL_G80_MSK_PWM1;
	outl((regval | CHG_PINSEL_G80_PWM1), CHG_PINSEL_G80);

	regval = inl(CHG_PULL_G88) & ~CHG_PULL_G88_MSK_PWM1;
	outl((regval | CHG_PULL_G88_PWM1), CHG_PULL_G88);

	/* stop clock */
	smu_clockgate_control(SMU_GCLKCTRL_PWM_PWCLK1, SMU_GCLKCTRL_CLOSE);
	smu_clockgate_control(SMU_GCLKCTRL_PWM_PCLK,   SMU_GCLKCTRL_CLOSE);

	/* select clock source */
	regval = inl(ASMU_DIVPWMPWCLK) & SMU_DIVPWMPWCLK_INIT_MSK;
	outl((regval | SMU_DIVPWMPWCLK_INIT), ASMU_DIVPWMPWCLK);

	/* start clock */
	smu_clockgate_control(SMU_GCLKCTRL_PWM_PWCLK1, SMU_GCLKCTRL_OPEN);
	smu_clockgate_control(SMU_GCLKCTRL_PWM_PCLK,   SMU_GCLKCTRL_OPEN);

	/* unreset PWM */
	smu_reset_control(SMU_RESETREQ_PWM, SMU_RESETREQ_UNRESET);

	/* setting PWM */
	outl(1, PWM_CH1_LOOP0);
	outl(0, PWM_INTENSET);
	outl((PWM_MODE|(config->backlight.pol<<2)), PWM_CH1_MODE);

	outl(PWM_DELAY, PWM_CH1_DELAY0);
	outl(PWM_TOTAL, PWM_CH1_TOTAL0);
	outl(PWM_LEDGE, PWM_CH1_LEDGE0);
	outl(PWM_TEDGE, PWM_CH1_TEDGE0);

	/* output PWM once */
	pwm_start();
	pwm_stop();
	udelay(10000);

	/* set BLON */
	gpio_set_value(LCDC_GIO_LCD_BLON, 0x1);
}

/*-----------------------------------------------------------------------------
 *  Function    : pwm_exit
 *	Abstract	: Exit PWM
 *  Input       : none
 *  Return      : none
 *	Note		: use only PWM1
 *----------------------------------------------------------------------------*/
void pwm_exit(void)
{
	/* clear BLON */
	gpio_set_value(LCDC_GIO_LCD_BLON, 0x0);

	pwm_stop();

	/* stop clock */
	smu_clockgate_control(SMU_GCLKCTRL_PWM_PWCLK1, SMU_GCLKCTRL_CLOSE);
	smu_clockgate_control(SMU_GCLKCTRL_PWM_PCLK,   SMU_GCLKCTRL_CLOSE);

	/* reset PWM */
	smu_reset_control(SMU_RESETREQ_PWM, SMU_RESETREQ_RESET);
}

/*-----------------------------------------------------------------------------
 *  Function    : pwm_set_value
 *  Abstruct    : Set PWM value
 *  Input       : config : LCDC configure
 *				: deg : 0-100
 *  Return      : none
 *----------------------------------------------------------------------------*/
int pwm_set_value(lcdc_config_t *config, int deg)
{
	u_int tedge;

	if ((deg < 0) || (deg > 100))
		return -1;

	tedge = (config->backlight.full * deg) / 100;

	pwm_stop();
	if (tedge > 0) {
		outl(tedge, PWM_CH1_TEDGE0);
		pwm_start();
	}

	return 0;
}

#endif	/* CONFIG_ICX */ 
