/*
 * smu.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "../reg.h"
#include "icx_io.h"
#include "smu.h"


/*******************/
/*@ Local routines */
/*******************/

/*-----------------------------------------------------------------------------
 *  Function    : smu_reset_wait
 *  Abstruct    : Wait to complete the reset
 *  Input       : dev	: device info
 *  Return      : 
 *----------------------------------------------------------------------------*/
static inline void smu_reset_wait(u_int dev)
{
	switch (dev) {
		case SMU_RESETREQ_LCD :
			udelay(1);
			break;
		case SMU_RESETREQ_IIC2 :
			udelay(100);
			break;
		default :
			break;
	}
}


/********************/
/*@ Global routines */
/********************/

/*-----------------------------------------------------------------------------
 *  Function    : smu_reset_control
 *  Abstruct    : Reset/Unreset a specified device
 *  Input       : dev		: device info
 *                0/1		: reset/unreset
 *  Return      : 0         : success
 *                -EINVAL   : invalid parameters
 *----------------------------------------------------------------------------*/
int smu_reset_control(u_int dev, u_char onoff)
{
	u_int mask, emask;
	u_int bit;

	bit = dev & SMU_REG_BIT_MASK;
	if (bit >= SMU_REG_MAX_BIT_LEN)
		return -EINVAL;

	if ((onoff != SMU_RESETREQ_RESET) && (onoff != SMU_RESETREQ_UNRESET))
		return -EINVAL;

	emask = 1 << bit;
	mask  = onoff << bit;

	switch (dev & SMU_REG_GROUP_MASK) {
		case SMU_RESETREQ0_GROUP:
			outl(emask, ASMU_RESETREQ0ENA);
			outl(mask,  ASMU_RESETREQ0);
			outl(0,     ASMU_RESETREQ0ENA);
			break;
		case SMU_RESETREQ1_GROUP:
			outl(emask, ASMU_RESETREQ1ENA);
			outl(mask,  ASMU_RESETREQ1);
			outl(0,     ASMU_RESETREQ1ENA);
			break;
		case SMU_RESETREQ2_GROUP:
			outl(emask, ASMU_RESETREQ2ENA);
			outl(mask,  ASMU_RESETREQ2);
			outl(0,     ASMU_RESETREQ2ENA);
			break;
		default :
			return -EINVAL;
	}

	smu_reset_wait(dev);

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : smu_clkctrl_on
 *  Abstruct    : Auto clock control ON
 *  Input       : dev		: device info
 *  Return      : 0         : success
 *                -EINVAL   : invalid parameters
 *----------------------------------------------------------------------------*/
int smu_clkctrl_on(u_int dev)
{
	u_int regval;
	u_int mask;
	u_int bit;

	bit = dev & SMU_REG_BIT_MASK;
	if (bit >= SMU_REG_MAX_BIT_LEN)
		return -EINVAL;

	mask = 1 << bit;

	switch (dev & SMU_REG_GROUP_MASK) {
		case SMU_AHBCLKCTRL0_GROUP:
			regval = inl(ASMU_AHBCLKCTRL0) | mask;
			outl(regval, ASMU_AHBCLKCTRL0);
			break;
		case SMU_APBCLKCTRL0_GROUP:
			regval = inl(ASMU_APBCLKCTRL0) | mask;
			outl(regval, ASMU_APBCLKCTRL0);
			break;
		case SMU_APBCLKCTRL1_GROUP:
			regval = inl(ASMU_APBCLKCTRL1) | mask;
			outl(regval, ASMU_APBCLKCTRL1);
			break;
		case SMU_CLKCTRL_GROUP:
			regval = inl(ASMU_CLKCTRL) | mask;
			outl(regval, ASMU_CLKCTRL);
			break;
		default :
			return -EINVAL;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : smu_clkctrl_off
 *  Abstruct    : Auto clock control OFF
 *  Input       : dev		: device info
 *  Return      : 0         : success
 *                -EINVAL   : invalid parameters
 *----------------------------------------------------------------------------*/
int smu_clkctrl_off(u_int dev)
{
	u_int regval;
	u_int mask;
	u_int bit;

	bit = dev & SMU_REG_BIT_MASK;
	if (bit >= SMU_REG_MAX_BIT_LEN)
		return -EINVAL;

	mask = 1 << bit;

	switch (dev & SMU_REG_GROUP_MASK) {
		case SMU_AHBCLKCTRL0_GROUP:
			regval = inl(ASMU_AHBCLKCTRL0) & ~mask;
			outl(regval, ASMU_AHBCLKCTRL0);
			break;
		case SMU_APBCLKCTRL0_GROUP:
			regval = inl(ASMU_APBCLKCTRL0) & ~mask;
			outl(regval, ASMU_APBCLKCTRL0);
			break;
		case SMU_APBCLKCTRL1_GROUP:
			regval = inl(ASMU_APBCLKCTRL1) & ~mask;
			outl(regval, ASMU_APBCLKCTRL1);
			break;
		case SMU_CLKCTRL_GROUP:
			regval = inl(ASMU_CLKCTRL) & ~mask;
			outl(regval, ASMU_CLKCTRL);
			break;
		default :
			return -EINVAL;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Function    : smu_clockgate_control
 *  Abstruct    : Open a gated clock control
 *  Input       : clkbit	: clock bit
 *                0/1		: close/open
 *  Return      : 0         : success
 *                -EINVAL   : invalid parameters
 *----------------------------------------------------------------------------*/
int smu_clockgate_control(u_int clkbit, u_char onoff)
{
	u_int mask, emask;
	u_int bit;

	bit = clkbit & SMU_REG_BIT_MASK;
	if (bit >= SMU_REG_MAX_BIT_LEN)
		return -EINVAL;

	if ((onoff != SMU_GCLKCTRL_CLOSE) && (onoff != SMU_GCLKCTRL_OPEN))
		return -EINVAL;

	emask = 1 << bit;
	mask  = onoff << bit;

	switch (clkbit & SMU_REG_GROUP_MASK) {
		case SMU_GCLKCTRL0_GROUP:
			outl(emask, ASMU_GCLKCTRL0ENA);
			outl(mask,  ASMU_GCLKCTRL0);
			outl(0,     ASMU_GCLKCTRL0ENA);
			break;
		case SMU_GCLKCTRL1_GROUP:
			outl(emask, ASMU_GCLKCTRL1ENA);
			outl(mask,  ASMU_GCLKCTRL1);
			outl(0,     ASMU_GCLKCTRL1ENA);
			break;
		case SMU_GCLKCTRL2_GROUP:
			outl(emask, ASMU_GCLKCTRL2ENA);
			outl(mask,  ASMU_GCLKCTRL2);
			outl(0,     ASMU_GCLKCTRL2ENA);
			break;
		case SMU_GCLKCTRL3_GROUP:
			outl(emask, ASMU_GCLKCTRL3ENA);
			outl(mask,  ASMU_GCLKCTRL3);
			outl(0,     ASMU_GCLKCTRL3ENA);
			break;
		default :
			return -EINVAL;
	}

	return 0;
}

#endif	/* CONFIG_ICX */ 
