/*
 * spi.c
 *
 * Copyright 2008,2009 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <common.h>

#ifdef CONFIG_ICX

#include <asm/errno.h>

#include "../reg.h"
#include "icx_io.h"
#include "smu.h"
#include "spi.h"


/*****************/
/*@ Local Config */
/*****************/

/* SPI mode:Master/Slave */
#define	SPI_M_S_MASTER		(0 << 1)
#define	SPI_M_S_SLAVE		(1 << 1)
#define	SPI_M_S				SPI_M_S_MASTER

/* SPI mode:DMA on/off */
#define	SPI_DMA_OFF			(0 << 0)
#define	SPI_DMA_ON			(1 << 0)
#define	SPI_DMA				SPI_DMA_OFF

/* SPI mode:NB */
#define	SPI_NB_MIN			8
#define	SPI_NB_MAX			32

/* SPI pol.CSW */
#define	SPI_POL_CSW			(7 << 12)

/* SPI pol.mask */
#define	SPI_CSW_MASK		0x0000F000
#define	SPI_POL_MASK		0x00000007


/****************/
/*@ Control SPI */
/****************/

/* SPx_CONTROL */
#define	SPx_CONTROL_START	0x00000001
#define	SPx_CONTROL_RD		0x00000004
#define	SPx_CONTROL_WRT		0x00000008
#define	SPx_CONTROL_RX_EMP	0x00000040
#define	SPx_CONTROL_RST		0x00000100

/* SPx_FFCLR */
#define	SPx_FFCLR_ALL		0x0000007F

/* SPx_RAW_STATUS */
#define	SPx_RAW_STATUS_RX_OVR_RAW	0x00000001
#define	SPx_RAW_STATUS_TX_UDR_RAW	0x00000002
#define	SPx_RAW_STATUS_TERR_RAW		0x00000010

#define	SPx_RAW_STATUS_TX_ALLERR_RAW	(	\
	SPx_RAW_STATUS_TX_UDR_RAW				\
)
#define	SPx_RAW_STATUS_RX_ALLERR_RAW	(	\
	SPx_RAW_STATUS_RX_OVR_RAW				\
  | SPx_RAW_STATUS_TERR_RAW					\
)
#define	SPx_RAW_STATUS_ALLERR_RAW		(	\
	SPx_RAW_STATUS_TX_ALLERR_RAW			\
  | SPx_RAW_STATUS_RX_ALLERR_RAW			\
)

/* Read/Write */
#define	SPI_READ			1
#define	SPI_WRITE			2
#define	SPI_RW				3


/***************/
/*@ PIN select */
/***************/

/* SP0 */
#define	CHG_PINSEL_SP0_SP0			0x00000000

#define	CHG_PULL1_MSK_SP0			0xFFFF0000
#define	CHG_PULL1_SP0				0x54550000

/* G48 */
#define	CHG_PINSEL_G48_MSK_SP0_CS1	0x00000003
#define	CHG_PINSEL_G48_SP0_CS1		0x00000001

/* G64 */
#define	CHG_PINSEL_G64_MSK_SP1		0x00FC0000
#define	CHG_PINSEL_G64_MSK_SP1_CS0	0x03000000
#define	CHG_PINSEL_G64_MSK_SP1_CS2	0x30000000
#define	CHG_PINSEL_G64_SP1			0x00A80000
#define	CHG_PINSEL_G64_SP1_CS0		0x02000000
#define	CHG_PINSEL_G64_SP1_CS2		0x20000000

#define CHG_PULL_G72_MSK_SP1		0x00007770
#define CHG_PULL_G72_MSK_SP1_CS0	0x00070000
#define CHG_PULL_G72_MSK_SP1_CS2	0x07000000
#define CHG_PULL_G72_SP1			0x00005450
#define CHG_PULL_G72_SP1_CS0		0x00050000
#define CHG_PULL_G72_SP1_CS2		0x05000000


/*******************************/
/*@ Structure(type definition) */
/*******************************/

/* SPI register info */
typedef struct {
	volatile u_int mode;		/* SPx_MODE				(0x00)	*/
	volatile u_int pol;			/* SPx_POL				(0x04)	*/
	volatile u_int control;		/* SPx_CONTROL			(0x08)	*/
	         u_int reserve1;	/* Reserve				(0x0C)	*/
	volatile u_int tx_data;		/* SPx_TX_DATA			(0x10)	*/
	volatile u_int rx_data;		/* SPx_RX_DATA			(0x14)	*/
	volatile u_int status;		/* SPx_STATUS			(0x18)	*/
	volatile u_int raw_status;	/* SPx_RAW_STATUS		(0x1C)	*/
	volatile u_int enset;		/* SPx_ENSET			(0x20)	*/
	volatile u_int enclr;		/* SPx_ENCLR			(0x24)	*/
	volatile u_int ffclr;		/* SPx_FFCLR			(0x28)	*/
	         u_int reserve2;	/* Reserve				(0x2C)	*/
	volatile u_int rx_fifo_p;	/* SPx_RX_FIFO_P		(0x30)	*/
	volatile u_int control2;	/* SPx_CONTROL2			(0x34)	*/
	volatile u_int tiecs;		/* SPx_TIECS			(0x38)	*/
} spi_regs_t;

/* SMU info */
typedef struct {
	u_int reset;				/* reset 						*/
	u_int sclk_div;				/* SPIx_SCLK_DIV				*/
	u_int pclk;					/* pclk							*/
	u_int sclk;					/* sclk							*/
	u_int pclk_ctrl;			/* pclk control					*/
	u_int sclk_ctrl;			/* sclk control					*/
} spi_smu_t;

/* SPI info */
typedef struct {
	spi_regs_t	*regs;			/* SPI register info			*/
	spi_smu_t	*smu;			/* SMU info						*/
	u_int		probe;			/* probed or not				*/ 
} spi_info_t;


/********************/
/*@ Local Variables */
/********************/

/* SMU register(SP0) */
static spi_smu_t spi_smu_sp0 = {
	.reset		= SMU_RESETREQ_SP0,
	.sclk_div	= ASMU_DIVSP0SCLK,
	.pclk		= SMU_GCLKCTRL_SP0_PCLK,
	.sclk		= SMU_GCLKCTRL_SP0_SCLK,
	.pclk_ctrl	= SMU_CLKCTRL_SP0_PCLK,
	.sclk_ctrl	= SMU_CLKCTRL_SP0_SCLK,
};

/* SMU register(SP1) */
static spi_smu_t spi_smu_sp1 = {
	.reset		= SMU_RESETREQ_SP1,
	.sclk_div	= ASMU_DIVSP1SCLK,
	.pclk		= SMU_GCLKCTRL_SP1_PCLK,
	.sclk		= SMU_GCLKCTRL_SP1_SCLK,
	.pclk_ctrl	= SMU_CLKCTRL_SP1_PCLK,
	.sclk_ctrl	= SMU_CLKCTRL_SP1_SCLK,
};

/* SPI info */
static spi_info_t spi_info[] = {
	{	/* SP0 */
		.regs	= (spi_regs_t *)SP0_BASE,
		.smu	= &spi_smu_sp0,
		.probe	= 0,
	},
	{	/* SP1 */
		.regs	= (spi_regs_t *)SP1_BASE,
		.smu	= &spi_smu_sp1,
		.probe	= 0,
	},
};

static u_int spi_clock_table[] = {
	1436,	/* 1434KHz */
	2867,	/* 2.867MHz */
	5734,	/* 5.734MHz */
	11469,	/* 11.469MHz */
	22938,	/* 22.938MHz */
	38229,	/* 38.229MHz */
};

static unsigned int spi_plldiv_table[] = {
	SMU_DIV160,		/* 1500KHz(1434KHz) */
	SMU_DIV80,		/* 3MHz   (2.87MHz) */
	SMU_DIV40,		/* 6MHz   (5.73MHz) */
	SMU_DIV20,		/* 12MHz  (11.47MHz) */
	SMU_DIV10,		/* 24MHz  (22.94MHz) */
	SMU_DIV6,		/* 48MHz  (38.23MHz) */
};


/*******************/
/*@ Local Routines */
/*******************/

/*-----------------------------------------------------------------------------
 *	Function	: spi_power_on
 *	Abstruct	: Power on SPI.
 *	Input		: spi	: SPI information
 *	Return		: 
 *----------------------------------------------------------------------------*/
static void spi_power_on(spi_info_t *spi)
{
	smu_clkctrl_off(spi->smu->sclk_ctrl);
	smu_clkctrl_off(spi->smu->pclk_ctrl);
	smu_clockgate_control(spi->smu->pclk, SMU_GCLKCTRL_OPEN);
	smu_clockgate_control(spi->smu->sclk, SMU_GCLKCTRL_OPEN);
	smu_reset_control(spi->smu->reset, SMU_RESETREQ_UNRESET);
	smu_clkctrl_on(spi->smu->pclk_ctrl);
	smu_clkctrl_on(spi->smu->sclk_ctrl);
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_power_off
 *	Abstruct	: Power off SPI.
 *	Input		: spi	: SPI information
 *	Return		: 
 *----------------------------------------------------------------------------*/
static void spi_power_off(spi_info_t *spi)
{
	smu_clkctrl_off(spi->smu->sclk_ctrl);
	smu_clkctrl_off(spi->smu->pclk_ctrl);
	smu_reset_control(spi->smu->reset, SMU_RESETREQ_RESET);
	smu_clockgate_control(spi->smu->pclk, SMU_GCLKCTRL_CLOSE);
	smu_clockgate_control(spi->smu->sclk, SMU_GCLKCTRL_CLOSE);
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_soft_reset
 *	Abstruct	: Reset SPI.
 *	Input		: spi		: SPI information
 *				: config	: SPI config
 *	Return		: 
 *----------------------------------------------------------------------------*/
static void spi_soft_reset(spi_info_t *spi, spi_config_t *config)
{
	spi->regs->control |= SPx_CONTROL_RST;
	udelay((1000 * 4 / spi_clock_table[config->clk]) + 1);
	spi->regs->control &= ~SPx_CONTROL_RST;
	udelay(200);
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_set_sclk_div
 *	Abstruct	: Set a divisor of SPI.
 *	Input		: spi		: SPI information
 *				  sclk		: clock frequency
 *	Return		: 0			: success
 *                -EINVAL	: invalid parameters
 *----------------------------------------------------------------------------*/
static int spi_set_sclk_div(spi_info_t *spi, u_int sclk)
{
	switch (sclk) {
		case SPI_CLK_1500KHZ :
		case SPI_CLK_3MHZ :
		case SPI_CLK_6MHZ :
		case SPI_CLK_12MHZ :
		case SPI_CLK_24MHZ :
		case SPI_CLK_48MHZ :
			break;
		default :
			return -EINVAL;
	}

	outl(spi_plldiv_table[sclk], spi->smu->sclk_div);

	return 0;
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_config
 *	Abstruct	: Configurate SPI.
 *	Input		: spi		: SPI information
 *				  config	: SPI config
 *                mode_flag : access mode(READ/WRITE/RW)
 *	Return		: 0			: success
 *                -EINVAL	: invalid parameters
 *  Note		: Not use TIECS
 *----------------------------------------------------------------------------*/
static int spi_config(spi_info_t *spi, spi_config_t *config, int mode_flag)
{
	u_int mode    = SPI_M_S | SPI_DMA;
	u_int csw_pol = 0;
	u_int pol     = 0;
	u_int pol_shift;
	u_int nb;

	/* set mode.nb */
	if (mode_flag == SPI_READ)
		nb = config->nbr;
	else if (mode_flag == SPI_WRITE)
		nb = config->nbw;
	else if (mode_flag == SPI_RW)
		nb = config->nbr + config->nbw;
	else
		return -EINVAL;

	if ((nb < SPI_NB_MIN) || (SPI_NB_MAX < nb))
		return -EINVAL;
	mode |= (nb - 1) << 8;

	/* set mode.cs and pol */
	switch (config->cs) {
		case SPI_CS_SEL_CS0 :
		case SPI_CS_SEL_CS1 :
		case SPI_CS_SEL_CS2 :
			mode |= config->cs;
			break;
		default :
			return -EINVAL;
	}

	pol_shift = (3 * (config->cs >> 4));
	pol = (config->pol & SPI_POL_MASK) << pol_shift;
	csw_pol = spi->regs->pol;
	csw_pol &= ~(SPI_CSW_MASK | (SPI_POL_MASK << pol_shift));
	csw_pol |= (SPI_POL_CSW | pol);

	/* set sclk */
	if (spi_set_sclk_div(spi, config->clk) < 0)
		return -EINVAL;

	/* set registers */
	spi->regs->mode = mode;
	spi->regs->pol  = csw_pol;
	
	/* software reset */
	spi_soft_reset(spi, config);

	return 0;
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_xferbytes
 *	Abstruct	: Send datas via SPI.
 *	Input		: spi		: SPI information
 *				  data		: data buffer
 *                mode_flag : access mode(READ/WRITE/RW)
 *	Return		: 0			: success
 *                -EIO		: failed to send
 *				  -EPERM	: invalid parameter
 *----------------------------------------------------------------------------*/
static int spi_xferbytes(spi_info_t *spi, u_int *data, int mode_flag)
{
	int   ret = 0;

	/* clear error status */
	spi->regs->ffclr = SPx_FFCLR_ALL;

	if (mode_flag == SPI_READ) {
		/* set the control */
		spi->regs->control = (SPx_CONTROL_RD | SPx_CONTROL_START);

		/* wait */
		while ((spi->regs->control & (SPx_CONTROL_START | SPx_CONTROL_RX_EMP)) != 0) {
			if ((spi->regs->raw_status & SPx_RAW_STATUS_RX_ALLERR_RAW) != 0) {
				ret = -EIO;
				break;
			}
		}

		/* recieve a data */
		*data = spi->regs->rx_data;
	}
	else if (mode_flag == SPI_WRITE) {
		/* send the data */
		spi->regs->tx_data = *data;
		spi->regs->control = (SPx_CONTROL_WRT | SPx_CONTROL_START);

		/* wait */
		while ((spi->regs->control & SPx_CONTROL_START) != 0) {
			if ((spi->regs->raw_status & SPx_RAW_STATUS_TX_ALLERR_RAW) != 0) {
				ret = -EIO;
				break;
			}
		}
	}
	else if (mode_flag == SPI_RW) {
		/* send the command */
		spi->regs->tx_data = *data;
		spi->regs->control = (SPx_CONTROL_RD | SPx_CONTROL_WRT | SPx_CONTROL_START);

		/* wait */
		while ((spi->regs->control & (SPx_CONTROL_START | SPx_CONTROL_RX_EMP)) != 0) {
			if ((spi->regs->raw_status & SPx_RAW_STATUS_ALLERR_RAW) != 0) {
				ret = -EIO;
				break;
			}
		}

		/* recieve a data */
		*data = spi->regs->rx_data;
	}
	else {
		ret = -EPERM;
	}

	return ret;
}


/********************/
/*@ Global Routines */
/********************/

/*-----------------------------------------------------------------------------
 *	Function	: spi_init
 *	Abstruct	: Initialize SPI.
 *	Input		: none
 *	Return		: none
 *  Note        : use SP0_CS0 (PowerIC)
 *                    SP1_CS0 (LCD)
 *                    SP1_CS2 (DAC)
 *                    SP0_CS1 (LCD : NEC reference board)
 *----------------------------------------------------------------------------*/
void spi_init(void)
{
	unsigned int pinsel;

	/* set PINSEL */
	outl(CHG_PINSEL_SP0_SP0, CHG_PINSEL_SP0);

	pinsel = inl(CHG_PINSEL_G64) & ~(CHG_PINSEL_G64_MSK_SP1	   |
									 CHG_PINSEL_G64_MSK_SP1_CS0|
									 CHG_PINSEL_G64_MSK_SP1_CS2);
	outl((pinsel				|
		  CHG_PINSEL_G64_SP1	|
		  CHG_PINSEL_G64_SP1_CS0|
		  CHG_PINSEL_G64_SP1_CS2), CHG_PINSEL_G64);

	/* set PULL */
	pinsel = inl(CHG_PULL1) & ~CHG_PULL1_MSK_SP0;
	outl((pinsel | CHG_PULL1_SP0), CHG_PULL1);

	pinsel = inl(CHG_PULL_G72) & ~(CHG_PULL_G72_MSK_SP1	   |
								   CHG_PULL_G72_MSK_SP1_CS0|
								   CHG_PULL_G72_MSK_SP1_CS2);
	outl((pinsel			  |
		  CHG_PULL_G72_SP1	  |
		  CHG_PULL_G72_SP1_CS0|
		  CHG_PULL_G72_SP1_CS2), CHG_PULL_G72);

#ifndef CONFIG_SONY_ICX
	pinsel = inl(CHG_PINSEL_G48) & ~CHG_PINSEL_G48_MSK_SP0_CS1;
	outl((pinsel | CHG_PINSEL_G48_SP0_CS1), CHG_PINSEL_G48);
#endif
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_probe
 *	Abstruct	: Probe SPI.
 *	Input		: config	: SPI config
 *	Return		: 0			: success
 *                -EINVAL	: invalid parameters
 *                -ENODEV	: no device
 *  Note        : use SP0_CS0 (PowerIC)
 *                    SP1_CS0 (LCD)
 *                    SP1_CS2 (DAC)
 *                    SP0_CS1 (LCD : NEC reference board)
 *----------------------------------------------------------------------------*/
int spi_probe(spi_config_t *config)
{
	spi_info_t *spi;

	/* check the parameters */
	if (config == NULL)
		return -EINVAL;

	if ((config->dev != SPI_DEV_SP0) && (config->dev != SPI_DEV_SP1))
		return -ENODEV;

	/* power on and configuration */
	spi = &spi_info[config->dev];

	if (spi->probe == 0) {
		spi_power_on(spi);

		/* software reset */
		spi_soft_reset(spi, config);

		spi->probe = 1;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_remove
 *	Abstruct	: Remove SPI.
 *	Input		: config	: SPI config
 *	Return		: 0			: success
 *                -EINVAL	: invalid parameters
 *                -ENODEV	: no device
 *----------------------------------------------------------------------------*/
int spi_remove(spi_config_t *config)
{
	spi_info_t *spi;

	/* check the parameters */
	if (config == NULL)
		return -EINVAL;

	if ((config->dev != SPI_DEV_SP0) && (config->dev != SPI_DEV_SP1))
		return -ENODEV;

	spi = &spi_info[config->dev];

	if (spi->probe != 0) {
		spi_power_off(spi);

		spi->probe = 0;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_write
 *	Abstruct	: Write a data to SPI.
 *	Input		: config	: SPI config
 *				  buf		: data buffer
 *				  count		: data length
 *	Return		: >= 0		: success
 *                -EINVAL	: invalid parameters
 *                -ENODEV	: no device
 *----------------------------------------------------------------------------*/
int spi_write(spi_config_t *config, u_int *buf, u_int count)
{
	spi_info_t *spi;
	u_int i;
	int   ret;

	/* check the parameters */
	if ((config == NULL) || (buf == NULL))
		return -EINVAL;

	/* set regs */
	if ((config->dev != SPI_DEV_SP0) && (config->dev != SPI_DEV_SP1))
		return -ENODEV;

	spi = &spi_info[config->dev];

	if (spi->probe == 0)
		return -EPERM;

	/* configuration */
	ret = spi_config(spi, config, SPI_WRITE);
	if (ret < 0)
		return ret;

	/* send data */
	ret = 0;
	for (i = 0; i < count; i++) {
		ret = spi_xferbytes(spi, &buf[i], SPI_WRITE);
		if (ret < 0) {
			spi_soft_reset(spi, config);
			break;
		}
	}
	if (ret == 0)
		ret = count;

	return ret;
}

/*-----------------------------------------------------------------------------
 *	Function	: spi_read
 *	Abstruct	: Read a data from SPI.
 *	Input		: config	: SPI config
 *				  cmd		: command buffer
 *				  buf		: data buffer
 *                flag		: read mode (1cycle/2cycle)
 *	Return		: 0			: success
 *                -EINVAL	: invalid parameters
 *                -ENODEV	: no device
 *	Note		: noly applied to read with command
 *----------------------------------------------------------------------------*/
int spi_read(spi_config_t *config, u_int *cmd, u_int *buf, u_int flag)
{
	spi_info_t *spi;
	u_int rw_data;
	int   ret;

	/* check the parameters */
	if ((config == NULL) || (buf == NULL))
		return -EINVAL;

	/* set regs */
	if ((config->dev != SPI_DEV_SP0) && (config->dev != SPI_DEV_SP1))
		return -ENODEV;

	spi = &spi_info[config->dev];

	if (spi->probe == 0)
		return -EPERM;

	if (flag == SPI_READ_1CYCLE) {
		/* send the command and recieve a data */
		ret = spi_config(spi, config, SPI_RW);
		if (ret == 0) {
			rw_data = *cmd << config->nbr;
			ret = spi_xferbytes(spi, &rw_data, SPI_RW);
			if (ret < 0) 
				spi_soft_reset(spi, config);
			*buf = rw_data;
		}
	}
	else if (flag == SPI_READ_2CYCLE) {
		/* send the command */
		ret = spi_config(spi, config, SPI_WRITE);
		if (ret == 0) {
			ret = spi_xferbytes(spi, cmd, SPI_WRITE);
			if (ret < 0) 
				spi_soft_reset(spi, config);
		}

		/* recieve a data */
		if (ret == 0) {
			ret = spi_config(spi, config, SPI_READ);
			if (ret == 0) {
				ret = spi_xferbytes(spi, buf, SPI_READ);
				if (ret < 0) 
					spi_soft_reset(spi, config);
			}
		}
	}
	else {
		ret =  -EPERM;
	}

	return ret;
}

#endif /* CONFIG_ICX */
